"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2025 Olivier Friard

This file is part of BORIS.

  BORIS is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  any later version.

  BORIS is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not see <http://www.gnu.org/licenses/>.

"""

# to run this file, use
# python -m boris.modifier_coding_map_creator

import binascii
import io
import json
from pathlib import Path
import re
import gui_utilities

from PySide6.QtCore import (
    Qt,
    Signal,
    QPoint,
    QByteArray,
    QBuffer,
    QIODevice,
    QLineF,
)
from PySide6.QtGui import QColor, QBrush, QMouseEvent, QPixmap, QIcon, QPen, QPolygon, QPolygonF, QAction
from PySide6.QtWidgets import (
    QGraphicsPolygonItem,
    QGraphicsEllipseItem,
    QGraphicsPixmapItem,
    QGraphicsLineItem,
    QMainWindow,
    QGraphicsView,
    QPushButton,
    QLabel,
    QHBoxLayout,
    QLineEdit,
    QSlider,
    QGraphicsScene,
    QWidget,
    QColorDialog,
    QVBoxLayout,
    QMessageBox,
    QInputDialog,
    QFileDialog,
    QApplication,
    QListWidget,
    QSplitter,
    QSpacerItem,
    QSizePolicy,
    QFrame,
)

from . import config as cfg
from . import dialog
from . import utilities as util

designColor = QColor(255, 0, 0, 128)  # red opacity: 50%
penWidth: int = 0
penStyle = Qt.NoPen
selectedBrush = QBrush()
selectedBrush.setStyle(Qt.SolidPattern)
selectedBrush.setColor(QColor(255, 255, 0, 255))


class ModifiersMapCreatorWindow(QMainWindow):
    closed = Signal()

    class View(QGraphicsView):
        """
        class for handling mousepress event in QGraphicsView
        """

        mousePress = Signal(QMouseEvent)

        def mousePressEvent(self, event):
            self.mousePress.emit(event)

        _start: int = 0
        elList: list = []
        points: list = []

        def __init__(self, parent):
            QGraphicsView.__init__(self, parent)
            self.setBackgroundBrush(QColor(128, 128, 128))
            self.setScene(QGraphicsScene(self))
            self.scene().update()

    bitmapFileName: str = ""
    mapName: str = ""
    fileName: str = ""
    flagNewArea: bool = False
    flag_map_changed: bool = False
    areasList: dict = {}
    polygonsList2: dict = {}
    areaColor = QColor("lime")

    def __init__(self):
        super(ModifiersMapCreatorWindow, self).__init__()

        self.pixmap = QPixmap()
        self.closedPolygon = None
        self.selectedPolygon = None

        self.setWindowTitle("BORIS - Modifiers map creator")

        self.newMapAction = QAction(QIcon(), "&New modifiers map", self)
        self.newMapAction.setShortcut("Ctrl+N")
        self.newMapAction.setStatusTip("Create a new modifiers map")
        self.newMapAction.triggered.connect(self.newMap)

        self.openMapAction = QAction(QIcon(), "&Open modifiers map", self)
        self.openMapAction.setShortcut("Ctrl+O")
        self.openMapAction.setStatusTip("Open a modifiers map")
        self.openMapAction.triggered.connect(self.openMap)

        self.saveMapAction = QAction(QIcon(), "&Save modifiers map", self)
        self.saveMapAction.setShortcut("Ctrl+S")
        self.saveMapAction.setStatusTip("Save modifiers map")
        self.saveMapAction.setEnabled(False)
        self.saveMapAction.triggered.connect(self.saveMap_clicked)

        self.saveAsMapAction = QAction(QIcon(), "Save modifiers map as", self)
        self.saveAsMapAction.setStatusTip("Save modifiers map as")
        self.saveAsMapAction.setEnabled(False)
        self.saveAsMapAction.triggered.connect(self.saveAsMap_clicked)

        self.mapNameAction = QAction(QIcon(), "&Modifiers map name", self)
        self.mapNameAction.setShortcut("Ctrl+M")
        self.mapNameAction.setStatusTip("Change modifiers map name")
        self.mapNameAction.setEnabled(False)
        self.mapNameAction.triggered.connect(self.mapName_clicked)

        self.exitAction = QAction(QIcon(), "&Close", self)
        self.exitAction.setStatusTip("Close modifiers map creator")
        self.exitAction.triggered.connect(self.close)

        menubar = self.menuBar()
        fileMenu = menubar.addMenu("&Modifiers Map creator")
        fileMenu.addAction(self.newMapAction)
        fileMenu.addAction(self.openMapAction)
        fileMenu.addAction(self.saveMapAction)
        fileMenu.addAction(self.saveAsMapAction)
        fileMenu.addSeparator()
        fileMenu.addAction(self.mapNameAction)
        fileMenu.addSeparator()
        fileMenu.addAction(self.exitAction)

        splitter1 = QSplitter(Qt.Vertical)

        self.view = self.View(self)
        self.view.mousePress.connect(self.viewMousePressEvent)
        splitter1.addWidget(self.view)

        vlayout_list = QVBoxLayout()
        vlayout_list.addWidget(QLabel("List of modifiers"))

        self.area_list = QListWidget(self)
        self.area_list.itemClicked.connect(self.area_list_item_click)
        vlayout_list.addWidget(self.area_list)
        w = QWidget()
        w.setLayout(vlayout_list)
        splitter1.addWidget(w)
        splitter1.setSizes([300, 100])
        splitter1.setStretchFactor(2, 8)

        hlayout_cmd = QHBoxLayout()

        self.btNewArea = QPushButton("New modifier", self)
        self.btNewArea.clicked.connect(self.newArea)
        self.btNewArea.setVisible(False)
        hlayout_cmd.addWidget(self.btNewArea)

        self.btSaveArea = QPushButton("Save modifier", self)
        self.btSaveArea.clicked.connect(self.saveArea)
        self.btSaveArea.setVisible(False)
        hlayout_cmd.addWidget(self.btSaveArea)

        self.btCancelAreaCreation = QPushButton("Cancel new modifier", self)
        self.btCancelAreaCreation.clicked.connect(self.cancelAreaCreation)
        self.btCancelAreaCreation.setVisible(False)
        hlayout_cmd.addWidget(self.btCancelAreaCreation)

        hlayout_cmd.addItem(QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum))

        self.btDeleteArea = QPushButton("Delete selected modifier", self)
        self.btDeleteArea.clicked.connect(self.deleteArea)
        self.btDeleteArea.setVisible(False)
        self.btDeleteArea.setEnabled(False)
        hlayout_cmd.addWidget(self.btDeleteArea)

        hlayout_area = QHBoxLayout()

        self.lb = QLabel("Modifier")
        self.lb.setVisible(False)
        hlayout_area.addWidget(self.lb)

        self.leAreaCode = QLineEdit(self)
        self.leAreaCode.setVisible(False)
        hlayout_area.addWidget(self.leAreaCode)

        self.btColor = QPushButton()
        self.btColor.clicked.connect(self.chooseColor)
        self.btColor.setVisible(False)
        self.btColor.setStyleSheet(f"QWidget {{ background-color:{self.areaColor.name()} }}")
        hlayout_area.addWidget(self.btColor)

        self.slAlpha = QSlider(Qt.Horizontal)
        self.slAlpha.setRange(20, 100)
        self.slAlpha.setValue(50)
        self.slAlpha.valueChanged.connect(self.slAlpha_changed)
        self.slAlpha.setVisible(False)
        hlayout_area.addWidget(self.slAlpha)

        self.slAlpha_changed(50)

        vlayout_frame = QVBoxLayout()
        vlayout_frame.addLayout(hlayout_cmd)
        vlayout_frame.addLayout(hlayout_area)
        vlayout_frame.addItem(QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding))

        frame = QFrame()
        frame.setFrameStyle(QFrame.Panel | QFrame.Plain)
        frame.setMinimumHeight(120)
        frame.setMaximumHeight(120)

        frame.setLayout(vlayout_frame)

        vlayout = QVBoxLayout()

        vlayout.addWidget(splitter1)
        vlayout.addWidget(frame)

        main_widget = QWidget(self)
        main_widget.setLayout(vlayout)
        self.setCentralWidget(main_widget)

        self.statusBar().showMessage("")

    def area_list_item_click(self, item):
        """
        select the polygon corresponding to the clicked area
        """

        if self.selectedPolygon:
            self.selectedPolygon.setPen(QPen(designColor, penWidth, penStyle, Qt.RoundCap, Qt.RoundJoin))
            self.selectedPolygon = None
            self.selectedPolygonMemBrush = None

        modifier_name = item.text()

        self.selectedPolygon = self.polygonsList2[item.text()]

        self.selectedPolygonMemBrush = self.selectedPolygon.brush()

        self.selectedPolygonAreaCode = modifier_name

        self.selectedPolygon.setPen(QPen(QColor(255, 0, 0, 255), 2, Qt.SolidLine, Qt.RoundCap, Qt.RoundJoin))

        self.leAreaCode.setText(modifier_name)
        for widget in (self.lb, self.leAreaCode, self.btSaveArea, self.btDeleteArea, self.btColor, self.slAlpha):
            widget.setVisible(True)

        self.btSaveArea.setText("Update modifier")
        self.btDeleteArea.setEnabled(True)

        self.areaColor = self.selectedPolygon.brush().color()
        self.btColor.setStyleSheet(f"QWidget {{background-color:{self.selectedPolygon.brush().color().name()}}}")

        self.slAlpha.setValue(int(self.areaColor.alpha() / 255 * 100))

    def get_current_alpha_value(self) -> str | None:
        """
        returns current alpha value (from button text)
        """
        match = re.search(r"(\d+)", self.btColor.text())

        if match:
            opacity_value = int(match.group(1))  # Convert the extracted string to an integer
            return opacity_value
        else:
            return None

    def slAlpha_changed(self, val):
        """
        opacity slider value changed
        """

        self.btColor.setText(f"Opacity: {val}")
        self.areaColor.setAlpha(int(val / 100 * 255))

        if self.selectedPolygon:
            self.selectedPolygon.setBrush(self.areaColor)

            # self.areasList[self.leAreaCode.text()]["color"] = self.areaColor.rgba()

        if self.closedPolygon:
            self.closedPolygon.setBrush(self.areaColor)

    def chooseColor(self):
        """
        area color button clicked
        """
        cd = QColorDialog()
        cd.setWindowFlags(Qt.WindowStaysOnTopHint)
        cd.setOptions(QColorDialog.ShowAlphaChannel | QColorDialog.DontUseNativeDialog)

        if cd.exec():
            self.areaColor = cd.currentColor()
            self.areaColor.setAlpha(int(self.get_current_alpha_value() / 100 * 255))

            self.btColor.setStyleSheet(f"QWidget {{background-color:{self.areaColor.name()}}}")

            if self.selectedPolygon:
                self.selectedPolygon.setBrush(self.areaColor)
                self.areasList[self.leAreaCode.text()]["color"] = self.areaColor.rgba()

            if self.closedPolygon:
                self.closedPolygon.setBrush(self.areaColor)

    def closeEvent(self, event):
        if self.flag_map_changed:
            response = dialog.MessageDialog(
                "BORIS - Modifiers map creator",
                "What to do about the current unsaved modifiers coding map?",
                [cfg.SAVE, cfg.DISCARD, cfg.CANCEL],
            )

            if response == cfg.SAVE:
                if not self.saveMap_clicked():
                    event.ignore()

            if response == cfg.CANCEL:
                event.ignore()
                return

        self.closed.emit()
        event.accept()

    def viewMousePressEvent(self, event):
        """
        check if area selected with mouse
        """

        def new_polygon():
            """
            create a newpolygon
            """
            newPolygon = QPolygonF()
            for p in self.view.points:
                newPolygon.append(QPoint(p[0], p[1]))

            # draw polygon a red polygon
            self.closedPolygon = QGraphicsPolygonItem(newPolygon)
            self.closedPolygon.setPen(QPen(designColor, penWidth, penStyle, Qt.RoundCap, Qt.RoundJoin))
            self.closedPolygon.setBrush(self.areaColor)
            self.view.scene().addItem(self.closedPolygon)

        def add_last_line():
            """
            add last line to start point
            """
            line = QGraphicsLineItem(
                QLineF(
                    self.view._start,
                    QPoint(self.view.points[0][0], self.view.points[0][1]),
                )
            )
            line.setPen(
                QPen(
                    designColor,
                    penWidth,
                    Qt.SolidLine,
                    Qt.RoundCap,
                    Qt.RoundJoin,
                )
            )

            self.view.scene().addItem(line)
            self.view.elList.append(line)

            self.statusBar().showMessage("Area completed")

        if not self.bitmapFileName:
            return

        self.btDeleteArea.setEnabled(False)

        # test = self.view.mapToScene(event.pos()).toPoint()
        test = self.view.mapToScene(event.position().toPoint()).toPoint()

        if test.x() < 0 or test.y() < 0 or test.x() > self.pixmap.size().width() or test.y() > self.pixmap.size().height():
            return

        if not self.flagNewArea:  # test clicked point for areas
            txt = ""

            # reset selected polygon to default pen
            if self.selectedPolygon:
                self.selectedPolygon.setPen(QPen(designColor, penWidth, penStyle, Qt.RoundCap, Qt.RoundJoin))
                self.selectedPolygon = None
                self.selectedPolygonMemBrush = None

            for idx, areaCode in enumerate(sorted(self.polygonsList2.keys())):
                for widget in (self.lb, self.leAreaCode, self.btSaveArea, self.btDeleteArea, self.btColor, self.slAlpha):
                    widget.setVisible(False)

                if self.polygonsList2[areaCode].contains(test):
                    txt += "," if txt else ""

                    txt += areaCode
                    self.selectedPolygon = self.polygonsList2[areaCode]
                    self.selectedPolygonAreaCode = areaCode

                    self.selectedPolygonMemBrush = self.selectedPolygon.brush()

                    self.selectedPolygon.setPen(
                        QPen(
                            QColor(255, 0, 0, 255),
                            2,
                            Qt.SolidLine,
                            Qt.RoundCap,
                            Qt.RoundJoin,
                        )
                    )

                    self.leAreaCode.setText(areaCode)
                    self.btDeleteArea.setEnabled(True)

                    # select area in list widget
                    item = self.area_list.item(idx)
                    self.area_list.setCurrentItem(item)

                    for widget in (self.lb, self.leAreaCode, self.btSaveArea, self.btDeleteArea, self.btColor, self.slAlpha):
                        widget.setVisible(True)

                    self.areaColor = self.selectedPolygon.brush().color()
                    self.btColor.setStyleSheet(f"QWidget {{background-color:{self.selectedPolygon.brush().color().name()} }}")

                    self.slAlpha.setValue(int(self.areaColor.alpha() / 255 * 100))

                    break

            if txt:
                self.statusBar().showMessage(f"Modifier{'s' if ',' in txt else ''}: {txt}")
            else:
                self.statusBar().showMessage("")

            if not self.selectedPolygon:
                self.leAreaCode.setVisible(False)
                self.btColor.setVisible(False)
                self.slAlpha.setVisible(False)
            return

        # right button: delete last line item
        if (event.buttons() & Qt.RightButton) and not self.closedPolygon:
            if self.view.points:
                self.view.points = self.view.points[0:-1]

            if self.view.points:
                self.view._start = QPoint(self.view.points[-1][0], self.view.points[-1][1])
            else:
                self.view._start = None

            # remove graphical elements
            if self.view.elList:
                self.view.scene().removeItem(self.view.elList[-1])
                self.view.elList = self.view.elList[0:-1]

        # middle button automatically close the polygon
        if (event.buttons() & Qt.MiddleButton) and not self.closedPolygon:
            add_last_line()
            new_polygon()

        # add line item
        if event.buttons() == Qt.LeftButton and not self.closedPolygon:
            if self.view._start:
                end = test

                # test is polygon is crossed
                if len(self.view.points) >= 3:
                    for idx, point in enumerate(self.view.points[:-2]):
                        if util.intersection(
                            self.view.points[idx],
                            self.view.points[idx + 1],
                            self.view.points[-1],
                            (int(end.x()), int(end.y())),
                        ):
                            QMessageBox.critical(self, "", "The polygon edges can not be intersected")
                            return

                # test if polygon closed (dist min 10 px)
                if abs(end.x() - self.view.points[0][0]) < 10 and abs(end.y() - self.view.points[0][1]) < 10:
                    add_last_line()
                    new_polygon()

                    return

                self.view.points.append((int(end.x()), int(end.y())))

                line = QGraphicsLineItem(QLineF(self.view._start, end))

                line.setPen(QPen(designColor, 2, Qt.SolidLine, Qt.RoundCap, Qt.RoundJoin))

                self.view.scene().addItem(line)
                self.view.elList.append(line)

                self.view._start = test

            else:  # first point
                self.view._start = test

                ellipse = QGraphicsEllipseItem(self.view._start.x(), self.view._start.y(), 3, 3)
                ellipse.setPen(QPen(designColor, 1, Qt.SolidLine, Qt.RoundCap, Qt.RoundJoin))

                brush = QBrush()
                brush.setStyle(Qt.SolidPattern)
                brush.setColor(designColor)
                ellipse.setBrush(brush)

                self.view.scene().addItem(ellipse)
                self.view.elList.append(ellipse)

                self.view.points.append((self.view._start.x(), self.view._start.y()))

    def mapName_clicked(self):
        """
        change map name
        """
        text, ok = QInputDialog.getText(
            self,
            "Modifiers map name",
            "Enter a name for the modifiers map",
            QLineEdit.Normal,
            self.mapName,
        )
        if ok:
            self.mapName = text
            self.setWindowTitle(f"{cfg.programName} - Modifiers map creator tool - {self.mapName}")

    def newMap(self):
        """
        create a new map
        """

        if self.flag_map_changed:
            response = dialog.MessageDialog(
                cfg.programName + " - Modifiers map creator",
                "What to do about the current unsaved coding map?",
                ["Save", "Discard", "Cancel"],
            )

            if response == "Save":
                if not self.saveMap_clicked():
                    return

            if response == "Cancel":
                return

        self.cancelMap()

        text, ok = QInputDialog.getText(self, "Map name", "Enter a name for the new map")
        if ok:
            self.mapName = text
        else:
            return

        if self.mapName == "":
            QMessageBox.critical(self, "", "You must define a name for the new map")
            return

        if self.mapName in ["areas", "bitmap"]:
            QMessageBox.critical(self, "", "This name is not allowed")
            return

        self.setWindowTitle(cfg.programName + " - Map creator tool - " + self.mapName)

        self.loadBitmap()

    def openMap(self):
        """
        load bitmap from data
        show it in view scene
        """
        if self.flag_map_changed:
            response = dialog.MessageDialog(
                cfg.programName + " - Map creator",
                "What to do about the current unsaved coding map?",
                (cfg.SAVE, cfg.DISCARD, cfg.CANCEL),
            )

            if response == cfg.SAVE:
                if not self.saveMap_clicked():
                    return

            if response == cfg.CANCEL:
                return

        fileName, _ = QFileDialog().getOpenFileName(
            self,
            "Open a coding map",
            "",
            "BORIS coding map (*.boris_map);;All files (*)",
        )

        if not fileName:
            return
        try:
            self.codingMap = json.loads(open(fileName, "r").read())
        except Exception:
            QMessageBox.critical(
                self,
                cfg.programName,
                f"The file {fileName} seems not a behaviors coding map...",
            )
            return

        self.cancelMap()

        self.mapName = self.codingMap["name"]

        self.setWindowTitle(cfg.programName + " - Map creator tool - " + self.mapName)

        self.bitmapFileName = True

        self.fileName = fileName

        self.areasList = self.codingMap["areas"]  # dictionary of dictionaries

        bitmapContent = binascii.a2b_base64(self.codingMap["bitmap"])

        self.pixmap.loadFromData(bitmapContent)

        self.btDeleteArea.setEnabled(False)

        self.view.setSceneRect(0, 0, self.pixmap.size().width(), self.pixmap.size().height())
        pixItem = QGraphicsPixmapItem(self.pixmap)
        pixItem.setPos(0, 0)
        self.view.scene().addItem(pixItem)

        for areaCode in self.areasList:
            points = self.areasList[areaCode]["geometry"]

            newPolygon = QPolygonF()
            for p in points:
                newPolygon.append(QPoint(p[0], p[1]))

            clr = QColor()
            clr.setRgba(self.areasList[areaCode]["color"])

            # draw polygon
            polygon = QGraphicsPolygonItem()

            polygon.setPolygon(newPolygon)

            polygon.setPen(QPen(clr, penWidth, penStyle, Qt.RoundCap, Qt.RoundJoin))

            polygon.setBrush(QBrush(clr, Qt.SolidPattern))

            self.view.scene().addItem(polygon)
            self.polygonsList2[areaCode] = polygon

        self.btNewArea.setVisible(True)

        self.saveMapAction.setEnabled(True)
        self.saveAsMapAction.setEnabled(True)
        self.mapNameAction.setEnabled(True)

        self.update_area_list()

        self.statusBar().showMessage('Click the "New modifier" button to create a new modifier')

    def saveMap(self) -> bool:
        """
        save the modifier coding map on file
        """
        if not self.fileName:
            return False
        # create dict with map name key
        mapDict = {"name": self.mapName}

        # add areas
        mapDict["areas"] = self.areasList

        # Save QPixmap to QByteArray via QBuffer.
        byte_array = QByteArray()
        buffer = QBuffer(byte_array)
        buffer.open(QIODevice.WriteOnly)
        self.pixmap.save(buffer, "PNG")

        string_io = io.BytesIO(byte_array)

        string_io.seek(0)

        # add bitmap
        mapDict["bitmap"] = binascii.b2a_base64(string_io.read()).decode("utf-8")

        with open(self.fileName, "w") as outfile:
            outfile.write(json.dumps(mapDict))

        self.flag_map_changed = False

        return True

    def saveAsMap_clicked(self):
        filters = "Modifiers map (*.boris_map);;All files (*)"

        fn = QFileDialog(self).getSaveFileName(self, "Save modifiers map as", "", filters)
        if type(fn) is tuple:
            self.fileName, _ = fn
        else:
            self.fileName = fn

        if self.fileName:
            # if os.path.splitext(self.fileName)[1] != ".boris_map":
            if Path(self.fileName).suffix != ".boris_map":
                self.fileName += ".boris_map"
            self.saveMap()

    def saveMap_clicked(self):
        if not self.fileName:
            fn = QFileDialog(self).getSaveFileName(
                self,
                "Save modifiers map",
                self.mapName + ".boris_map",
                "BORIS MAP (*.boris_map);;All files (*)",
            )
            if type(fn) is tuple:
                self.fileName, _ = fn
            else:
                self.fileName = fn

            if self.fileName and Path(self.fileName).suffix() != ".boris_map":
                self.fileName += ".boris_map"

        if self.fileName:
            return self.saveMap()

        return False

    def newArea(self):
        """
        create a new area
        """
        if not self.bitmapFileName:
            QMessageBox.critical(self, cfg.programName, "An image must be loaded before to define areas")
            return

        if self.selectedPolygon:
            self.selectedPolygon.setPen(QPen(designColor, penWidth, penStyle, Qt.RoundCap, Qt.RoundJoin))
            self.selectedPolygon = None

        self.flagNewArea = True
        self.leAreaCode.clear()

        for widget in (
            self.btSaveArea,
            self.btCancelAreaCreation,
            self.lb,
            self.leAreaCode,
            self.btColor,
            self.slAlpha,
        ):
            widget.setVisible(True)

        self.btSaveArea.setText("Save modifier")

        self.btNewArea.setVisible(False)
        self.btDeleteArea.setVisible(False)

        self.statusBar().showMessage(
            "Draw a polygon by clicking the vertices (right-click will delete the last point, middle-click will close the polygon)"
        )

    def saveArea(self):
        """
        save the new modifier in the list
        """

        # check if not allowed character
        for c in cfg.CHAR_FORBIDDEN_IN_MODIFIERS:
            if c in self.leAreaCode.text():
                QMessageBox.critical(
                    self,
                    cfg.programName,
                    f"The modifier contains a character that is not allowed <b>{cfg.CHAR_FORBIDDEN_IN_MODIFIERS}</b>.",
                )
                return

        # check if modification
        if "save" in self.btSaveArea.text().lower():
            if not self.closedPolygon:
                QMessageBox.critical(
                    self,
                    cfg.programName,
                    "You must close your area before saving it.\nThe last vertex must correspond to the first one.",
                )

            if len(self.view.points) < 3:
                QMessageBox.critical(self, cfg.programName, "You must define a closed area")
                return

            # check if no modifier name
            if not self.leAreaCode.text():
                QMessageBox.critical(self, cfg.programName, "You must define a code for the new modifier")
                return

            # check if modifier name already used
            if self.leAreaCode.text() in self.areasList:
                QMessageBox.critical(self, cfg.programName, "The modifier name is already in use")
                return

            # create polygon
            newPolygon = QPolygon()
            for p in self.view.points:
                newPolygon.append(QPoint(p[0], p[1]))

            self.areasList[self.leAreaCode.text()] = {
                "geometry": self.view.points,
                "color": self.areaColor.rgba(),
            }

            # remove all lines
            for line in self.view.elList:
                self.view.scene().removeItem(line)

            # draw polygon
            self.closedPolygon.setBrush(QBrush(self.areaColor, Qt.SolidPattern))
            self.polygonsList2[self.leAreaCode.text()] = self.closedPolygon
            self.closedPolygon = None
            self.view._start = 0
            self.view.points = []
            self.view.elList = []
            self.flagNewArea = False

            self.statusBar().showMessage("New modifier saved", 5000)

        else:  # modification
            if self.leAreaCode.text() not in self.polygonsList2:
                self.polygonsList2[self.leAreaCode.text()] = self.polygonsList2.pop(self.area_list.currentItem().text())
                self.polygonsList2[self.leAreaCode.text()].setBrush(self.areaColor)

            self.statusBar().showMessage("Modifier modified", 5000)

        for widget in (
            self.btSaveArea,
            self.btCancelAreaCreation,
            self.lb,
            self.leAreaCode,
            self.btColor,
            self.slAlpha,
            self.btDeleteArea,
            self.btNewArea,
        ):
            widget.setVisible(False)

        self.btNewArea.setVisible(True)
        self.leAreaCode.setText("")
        self.update_area_list()
        self.flag_map_changed = True

    def cancelAreaCreation(self):
        if self.closedPolygon:
            self.view.scene().removeItem(self.closedPolygon)
            self.closedPolygon = None

        # remove all lines
        for line in self.view.elList:
            self.view.scene().removeItem(line)

        self.view.elList = []

        self.view._start = 0
        self.view.points = []
        self.flagNewArea = False
        self.btCancelAreaCreation.setVisible(False)
        self.btDeleteArea.setVisible(True)
        self.btSaveArea.setVisible(False)
        self.lb.setVisible(False)

        self.btColor.setVisible(False)
        self.slAlpha.setVisible(False)
        self.btNewArea.setVisible(True)

        self.leAreaCode.setVisible(False)
        self.leAreaCode.setText("")

    def update_area_list(self):
        """
        update the area list widget
        """
        self.area_list.clear()

        print(f"{self.polygonsList2=}")

        for modifier_name in sorted(self.polygonsList2.keys()):
            self.area_list.addItem(modifier_name)

    def deleteArea(self):
        """
        remove selected area from map
        """

        if self.selectedPolygon:
            self.view.scene().removeItem(self.selectedPolygon)

            if self.selectedPolygonAreaCode:
                self.view.scene().removeItem(self.polygonsList2[self.selectedPolygonAreaCode])
                del self.polygonsList2[self.selectedPolygonAreaCode]
                del self.areasList[self.selectedPolygonAreaCode]

            self.selectedPolygonAreaCode = None
            self.selectedPolygon = None

            self.flag_map_changed = True

        self.view.elList = []

        self.view._start = 0
        self.view.points = []
        self.flagNewArea = False

        for widget in (self.btSaveArea, self.lb, self.btColor, self.slAlpha, self.leAreaCode):
            widget.setVisible(False)

        self.btNewArea.setVisible(True)

        self.leAreaCode.setText("")
        self.statusBar().showMessage("")

        self.update_area_list()

    def cancelMap(self):
        """
        remove current map
        """
        self.flagNewArea = False
        self.areasList = {}
        self.polygonsList2 = {}
        self.view.scene().clear()
        self.btDeleteArea.setVisible(False)
        self.btNewArea.setVisible(False)
        self.saveMapAction.setEnabled(False)
        self.saveAsMapAction.setEnabled(False)
        self.mapNameAction.setEnabled(False)
        self.statusBar().showMessage("")
        self.flag_map_changed = False

    def loadBitmap(self):
        """
        load bitmap as background for coding map
        """

        fileName, _ = QFileDialog().getOpenFileName(self, "Load bitmap", "", "bitmap files (*.png *.jpg);;All files (*)")

        if not fileName:
            return
        self.bitmapFileName = fileName

        self.pixmap.load(self.bitmapFileName)

        # scale image
        """
        if self.pixmap.size().width() > maxSize or self.pixmap.size().height() > maxSize:
            self.pixmap = self.pixmap.scaled(maxSize, maxSize, Qt.KeepAspectRatio)
            QMessageBox.information(
                self,
                cfg.programName,
                "The bitmap was resized to %d x %d pixels\nThe original file was not modified"
                % (self.pixmap.size().width(), self.pixmap.size().height()),
            )
        """

        self.view.setSceneRect(0, 0, self.pixmap.size().width(), self.pixmap.size().height())
        pixitem = QGraphicsPixmapItem(self.pixmap)
        pixitem.setPos(0, 0)
        self.view.scene().addItem(pixitem)

        self.btNewArea.setVisible(True)

        self.saveMapAction.setEnabled(True)
        self.saveAsMapAction.setEnabled(True)
        self.mapNameAction.setEnabled(True)

        self.statusBar().showMessage("""Click "New modifier" to create a new modifier""")

        self.flag_map_changed = True


if __name__ == "__main__":
    import sys

    app = QApplication(sys.argv)
    window = ModifiersMapCreatorWindow()

    gui_utilities.resize_center(app, window, cfg.CODING_MAP_RESIZE_W, cfg.CODING_MAP_RESIZE_H)

    window.show()
    sys.exit(app.exec())



"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2025 Olivier Friard

This file is part of BORIS.

  BORIS is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  any later version.

  BORIS is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not see <http://www.gnu.org/licenses/>.

"""

from PySide6.QtCore import Qt
from PySide6.QtWidgets import QDialog

from . import config as cfg
from .param_panel_ui import Ui_Dialog
from . import dialog


class Param_panel(QDialog, Ui_Dialog):
    def __init__(self, parent=None):
        super().__init__()
        self.setupUi(self)

        self.media_duration = None

        # insert duration widget for time offset
        self.start_time = dialog.get_time_widget(0)
        self.horizontalLayout.insertWidget(1, self.start_time)
        self.end_time = dialog.get_time_widget(0)
        self.horizontalLayout_6.insertWidget(1, self.end_time)

        self.pbSelectAllSubjects.clicked.connect(lambda: self.subjects_button_clicked("select all"))
        self.pbUnselectAllSubjects.clicked.connect(lambda: self.subjects_button_clicked("unselect all"))
        self.pbReverseSubjectsSelection.clicked.connect(lambda: self.subjects_button_clicked("reverse selection"))

        self.pbSelectAllBehaviors.clicked.connect(lambda: self.behaviors_button_clicked("select all"))
        self.pbUnselectAllBehaviors.clicked.connect(lambda: self.behaviors_button_clicked("unselect all"))
        self.pbReverseBehaviorsSelection.clicked.connect(lambda: self.behaviors_button_clicked("reverse selection"))

        self.pbOK.clicked.connect(self.ok)
        self.pbCancel.clicked.connect(self.reject)

        self.lwBehaviors.itemClicked.connect(self.behavior_item_clicked)

        self.rb_observed_events.setChecked(True)

        self.rb_media_duration.clicked.connect(lambda: self.rb_time_interval_selection(cfg.TIME_FULL_OBS))
        self.rb_observed_events.clicked.connect(lambda: self.rb_time_interval_selection(cfg.TIME_EVENTS))
        self.rb_user_defined.clicked.connect(lambda: self.rb_time_interval_selection(cfg.TIME_ARBITRARY_INTERVAL))
        self.rb_obs_interval.clicked.connect(lambda: self.rb_time_interval_selection(cfg.TIME_OBS_INTERVAL))

        self.cbIncludeModifiers.stateChanged.connect(self.cb_exclude_non_coded_modifiers_visibility)

        self.cb_exclude_non_coded_modifiers.setVisible(False)

    def cb_exclude_non_coded_modifiers_visibility(self):
        """
        set visibility of cb_exclude_non_coded_modifiers
        """
        self.cb_exclude_non_coded_modifiers.setEnabled(self.cbIncludeModifiers.isChecked())

    def rb_time_interval_selection(self, button):
        """
        select the time interval for operation
        """
        if button == cfg.TIME_ARBITRARY_INTERVAL:
            self.frm_time_interval.setEnabled(True)
            self.frm_time_interval.setVisible(True)

        elif button == cfg.TIME_EVENTS and len(self.selectedObservations) == 1:
            self.start_time.set_time(self.start_coding)
            self.end_time.set_time(self.end_coding)
            self.frm_time_interval.setEnabled(False)
            self.frm_time_interval.setVisible(True)

        elif button == cfg.TIME_FULL_OBS and len(self.selectedObservations) == 1 and self.media_duration is not None:
            self.start_time.set_time(0)
            self.end_time.set_time(self.media_duration)
            self.frm_time_interval.setEnabled(False)
            self.frm_time_interval.setVisible(True)

        elif button == cfg.TIME_OBS_INTERVAL:
            if not ((self.start_interval is None) or self.start_interval.is_nan()):
                # Set start_time and end_time widgets values even if it is not shown with
                # more than 1 observation as some analyses might use it (eg: advanced event filtering)

                print(f"{self.end_interval=}")

                end_interval = self.end_interval if self.end_interval != 0 else self.media_duration

                print(f"{end_interval=}")

                self.start_time.set_time(self.start_interval)
                self.end_time.set_time(end_interval)
                self.frm_time_interval.setEnabled(False)
                if len(self.selectedObservations) == 1:
                    self.frm_time_interval.setVisible(True)

        else:
            self.frm_time_interval.setVisible(False)

    def subjects_button_clicked(self, command):
        for idx in range(self.lwSubjects.count()):
            cb = self.lwSubjects.itemWidget(self.lwSubjects.item(idx))
            if command == "select all":
                cb.setChecked(True)
            if command == "unselect all":
                cb.setChecked(False)
            if command == "reverse selection":
                cb.setChecked(not cb.isChecked())

    def behaviors_button_clicked(self, command):
        for idx in range(self.lwBehaviors.count()):
            if self.lwBehaviors.item(idx).data(33) != "category":
                if command == "select all":
                    self.lwBehaviors.item(idx).setCheckState(Qt.Checked)

                if command == "unselect all":
                    self.lwBehaviors.item(idx).setCheckState(Qt.Unchecked)

                if command == "reverse selection":
                    if self.lwBehaviors.item(idx).checkState() == Qt.Checked:
                        self.lwBehaviors.item(idx).setCheckState(Qt.Unchecked)
                    else:
                        self.lwBehaviors.item(idx).setCheckState(Qt.Checked)

    def ok(self):
        selectedSubjects = []
        for idx in range(self.lwSubjects.count()):
            cb = self.lwSubjects.itemWidget(self.lwSubjects.item(idx))
            if cb.isChecked():
                selectedSubjects.append(cb.text())
        self.selectedSubjects = selectedSubjects

        selectedBehaviors = []
        for idx in range(self.lwBehaviors.count()):
            if self.lwBehaviors.item(idx).checkState() == Qt.Checked:
                selectedBehaviors.append(self.lwBehaviors.item(idx).text())
        self.selectedBehaviors = selectedBehaviors

        self.accept()

    def behavior_item_clicked(self, item):
        """
        check / uncheck behaviors belonging to the clicked category
        """

        if item.data(33) == "category":
            category = item.data(34)
            for i in range(self.lwBehaviors.count()):
                if self.lwBehaviors.item(i).data(34) == category and self.lwBehaviors.item(i).data(33) != "category":
                    if item.data(35):
                        self.lwBehaviors.item(i).setCheckState(Qt.Unchecked)
                    else:
                        self.lwBehaviors.item(i).setCheckState(Qt.Checked)

            item.setData(35, not item.data(35))

    def extract_observed_behaviors(self, selected_observations, selected_subjects):
        """
        extract unique behaviors codes from obs_id observation and selected subjects
        """

        observed_behaviors = []

        # extract events from selected observations
        all_events = [self.pj[cfg.OBSERVATIONS][x][cfg.EVENTS] for x in self.pj[cfg.OBSERVATIONS] if x in selected_observations]

        for events in all_events:
            for event in events:
                if event[cfg.EVENT_SUBJECT_FIELD_IDX] in selected_subjects or (
                    not event[cfg.EVENT_SUBJECT_FIELD_IDX] and cfg.NO_FOCAL_SUBJECT in selected_subjects
                ):
                    observed_behaviors.append(event[cfg.EVENT_BEHAVIOR_FIELD_IDX])

        # remove duplicate
        return list(set(observed_behaviors))

    def cb_changed(self):
        selected_subjects: list = []
        for idx in range(self.lwSubjects.count()):
            cb = self.lwSubjects.itemWidget(self.lwSubjects.item(idx))
            if cb and cb.isChecked():
                selected_subjects.append(cb.text())

        observed_behaviors = self.extract_observed_behaviors(self.selectedObservations, selected_subjects)

        for idx in range(self.lwBehaviors.count()):
            if self.lwBehaviors.item(idx).data(33) != "category":
                if self.lwBehaviors.item(idx).text() in observed_behaviors:
                    self.lwBehaviors.item(idx).setCheckState(Qt.Checked)
                else:
                    self.lwBehaviors.item(idx).setCheckState(Qt.Unchecked)



"""
module for testing observation GUI


pytest -s -vv test_observation_gui.py
"""


import sys
import os
from PyQt5.QtCore import *

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from boris import observation
from boris import config

def test_no_media_loaded(qtbot):

    w = observation.Observation("/tmp")
    #w.show()

    qtbot.addWidget(w)

    def handle_dialog():
        qtbot.keyClick(w.qm, Qt.Key_Enter)

    QTimer.singleShot(1000, handle_dialog)

    qtbot.mouseClick(w.pbSave, Qt.LeftButton)

    assert w.state == "refused"


def test_no_obs_id(qtbot):

    w = observation.Observation("/tmp")
    #w.show()
    qtbot.addWidget(w)
    w.ffmpeg_bin = "ffmpeg"

    media_file = "files/geese1.mp4"
    w.check_media("1", media_file, True)
    w.add_media_to_listview("1", media_file)

    def handle_dialog():
        qtbot.keyClick(w.qm, Qt.Key_Enter)
        #qtbot.mouseClick(w.qm.Ok, Qt.LeftButton)

    QTimer.singleShot(1000, handle_dialog)

    qtbot.mouseClick(w.pbSave, Qt.LeftButton)

    assert w.state == "refused"


def test_file_not_media(qtbot):
    """
    test if the loaded file is a media file or not
    """

    w = observation.Observation("/tmp")
    w.show()
    qtbot.addWidget(w)
    w.mode = "new"
    w.pj = config.EMPTY_PROJECT
    w.ffmpeg_bin = "ffmpeg"

    w.leObservationId.setText("test")
    media_file = "files/test.boris"
    r, msg = w.check_media("1", media_file, True)
    assert r == False


def test_players_in_crescent_order(qtbot):
    """
    test if players are used in crescent order
    """

    w = observation.Observation("/tmp")
    #w.show()
    qtbot.addWidget(w)
    #w.mode = "new"
    w.pj = config.EMPTY_PROJECT
    w.ffmpeg_bin = "ffmpeg"

    #w.leObservationId.setText("test")
    media_file1 = "files/geese1.mp4"
    w.check_media("1", media_file1, True)
    #w.add_media_to_listview("1", media_file1)

    media_file2 = "files/geese1.mp4"
    w.check_media("1", media_file2, True)
    w.twVideo1.cellWidget(1, 0).setCurrentIndex(2)

    def handle_dialog():
        qtbot.keyClick(w.qm, Qt.Key_Enter)
        #qtbot.mouseClick(w.qm.Ok, Qt.LeftButton)

    QTimer.singleShot(1000, handle_dialog)

    qtbot.mouseClick(w.pbSave, Qt.LeftButton)

    assert w.state == "refused"


def test_ok(qtbot):

    w = observation.Observation("/tmp")
    #w.show()
    qtbot.addWidget(w)
    w.mode = "new"
    w.pj = config.EMPTY_PROJECT
    w.ffmpeg_bin = "ffmpeg"

    media_file = "files/geese1.mp4"
    w.leObservationId.setText("test")
    w.check_media("1", media_file, True)
    w.add_media_to_listview("1", media_file)

    qtbot.mouseClick(w.pbSave, Qt.LeftButton)

    assert w.state == "accepted"

    assert w.pj == {'time_format': 'hh:mm:ss', 'project_date': '', 'project_name': '', 'project_description': '',
                    'project_format_version': config.project_format_version,
                    'subjects_conf': {}, 'behaviors_conf': {}, 'observations': {},
                    'behavioral_categories': [],
                    'independent_variables': {}, 'coding_map': {}, 'behaviors_coding_map': [], 'converters': {}}


def test_cancel(qtbot):

    w = observation.Observation("/tmp")
    #w.show()
    qtbot.addWidget(w)
    w.mode = "new"
    w.pj = config.EMPTY_PROJECT
    w.ffmpeg_bin = "ffmpeg"

    media_file = "files/geese1.mp4"
    w.leObservationId.setText("test")
    w.check_media("1", media_file, True)
    w.add_media_to_listview("1", media_file)

    qtbot.mouseClick(w.pbCancel, Qt.LeftButton)

    assert w.pj == config.EMPTY_PROJECT



def test_extract_wav_from_video(qtbot):

    try:
        os.remove("/tmp/geese1.mp4.wav")
    except Exception:
        pass

    w = observation.Observation("/tmp")
    #w.show()
    qtbot.addWidget(w)
    w.mode = "new"
    w.pj = config.EMPTY_PROJECT
    w.ffmpeg_bin = "ffmpeg"
    w.ffmpeg_cache_dir = "/tmp"

    media_file = "files/geese1.mp4"
    w.leObservationId.setText("test")
    w.check_media("1", media_file, True)
    w.add_media_to_listview("1", media_file)

    w.cbVisualizeSpectrogram.setChecked(True)
    w.extract_wav()

    assert os.path.isfile("/tmp/geese1.mp4.wav")


def test_extract_wav_from_wav(qtbot):

    w = observation.Observation("/tmp")

    #w.show()
    qtbot.addWidget(w)
    w.mode = "new"
    w.pj = config.EMPTY_PROJECT
    w.ffmpeg_bin = "ffmpeg"
    w.ffmpeg_cache_dir = "/tmp"

    media_file = "files/test.wav"
    w.leObservationId.setText("test")
    w.check_media("1", media_file, True)
    w.add_media_to_listview("1", media_file)

    w.cbVisualizeSpectrogram.setChecked(True)
    w.extract_wav()

    assert os.path.isfile("/tmp/test.wav.wav")




"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2025 Olivier Friard

This file is part of BORIS.

  BORIS is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  any later version.

  BORIS is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not see <http://www.gnu.org/licenses/>.

"""

from PySide6.QtWidgets import QFileDialog

from . import config as cfg
from . import utilities as util
from . import dialog
from . import project_functions
from . import utilities as util


def get_info(self) -> None:
    """
    show info about media file (current media file if an observation is opened)
    """

    def media_analysis_str(ffmpeg_bin: str, media_full_path: str) -> str:
        r = util.accurate_media_analysis(ffmpeg_bin, media_full_path)

        if "error" in r:
            ffmpeg_output = f"File path: {media_full_path}<br><br>{r['error']}<br><br>"
        else:
            ffmpeg_output = f"<br><b>{r['analysis_program']} analysis</b><br>"

            ffmpeg_output += (
                f"File path: <b>{media_full_path}</b><br><br>"
                f"Duration: {r['duration']} seconds ({util.convertTime(self.timeFormat, r['duration'])})<br>"
                f"FPS: {r['fps']}<br>"
                f"Resolution: {r['resolution']} pixels<br>"
                f"Format long name: {r.get('format_long_name', cfg.NA)}<br>"
                f"Creation time: {r.get('creation_time', cfg.NA)}<br>"
                f"Number of frames: {r['frames_number']}<br>"
                f"Bitrate: {util.smart_size_format(r['bitrate'])}   <br>"
                f"Has video: {r['has_video']}<br>"
                f"Has audio: {r['has_audio']}<br>"
                f"File size: {util.smart_size_format(r.get('file size', cfg.NA))}<br>"
                f"Video codec: {r.get('video_codec', cfg.NA)}<br>"
                f"Audio codec: {r.get('audio_codec', cfg.NA)}<br>"
            )

        return ffmpeg_output

    if self.observationId and self.playerType == cfg.MEDIA:
        tot_output: str = ""

        for i, dw in enumerate(self.dw_player):
            if not (
                str(i + 1) in self.pj[cfg.OBSERVATIONS][self.observationId][cfg.FILE]
                and self.pj[cfg.OBSERVATIONS][self.observationId][cfg.FILE][str(i + 1)]
            ):
                continue

            mpv_output = (
                "<b>MPV information</b><br>"
                f"Duration: {dw.player.duration} seconds ({util.seconds2time(dw.player.duration)})<br>"
                # "Position: {} %<br>"
                f"FPS: {dw.player.container_fps}<br>"
                # "Rate: {}<br>"
                f"Resolution: {dw.player.width}x{dw.player.height} pixels<br>"
                # "Scale: {}<br>"
                f"Video format: {dw.player.video_format}<br>"
                # "State: {}<br>"
                # "Media Resource Location: {}<br>"
                # "File name: {}<br>"
                # "Track: {}/{}<br>"
                f"Number of media in media list: {dw.player.playlist_count}<br>"
                f"Current time position: {dw.player.time_pos}<br>"
                f"Aspect ratio: {round(dw.player.width / dw.player.height, 3)}<br>"
                # "is seekable? {}<br>"
                # "has_vout? {}<br>"
            )

            # FFmpeg/FFprobe analysis
            ffmpeg_output: str = ""
            for file_path in self.pj[cfg.OBSERVATIONS][self.observationId][cfg.FILE][str(i + 1)]:
                media_full_path = project_functions.full_path(file_path, self.projectFileName)
                ffmpeg_output += media_analysis_str(self.ffmpeg_bin, media_full_path)

            ffmpeg_output += f"<br>Total duration: {sum(self.dw_player[i].media_durations) / 1000} ({util.convertTime(self.timeFormat, sum(self.dw_player[i].media_durations) / 1000)})"

            tot_output += mpv_output + ffmpeg_output + "<br><hr>"

    else:  # no open observation
        file_paths, _ = QFileDialog().getOpenFileNames(self, "Select a media file", "", "Media files (*)")
        if not file_paths:
            return

        tot_output: str = ""
        for file_path in file_paths:
            tot_output += media_analysis_str(self.ffmpeg_bin, file_path)

    self.results = dialog.Results_dialog()
    self.results.setWindowTitle(f"{cfg.programName} - Media file information")
    self.results.ptText.appendHtml(tot_output)
    self.results.show()



"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2025 Olivier Friard


  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
  MA 02110-1301, USA.

"""

import logging
import copy
import time
from decimal import Decimal as dec
from decimal import InvalidOperation
from decimal import ROUND_DOWN
from typing import Union


from . import config as cfg
from . import utilities as util
from . import dialog
from . import select_subj_behav
from . import select_modifiers
from . import write_event
from .edit_event import DlgEditEvent, EditSelectedEvents

from PySide6.QtWidgets import QMessageBox, QInputDialog, QLineEdit, QAbstractItemView, QApplication
from PySide6.QtCore import QTime, Qt
from PySide6.QtGui import QClipboard


def add_event(self):
    """
    manually add event to observation
    """

    if not self.observationId:
        self.no_observation()
        return

    if self.pause_before_addevent:
        # pause media
        if self.pj[cfg.OBSERVATIONS][self.observationId][cfg.TYPE] == cfg.MEDIA and self.playerType == cfg.MEDIA:
            memState = self.is_playing()
            if memState:
                self.pause_video()

    if not self.pj[cfg.ETHOGRAM]:
        QMessageBox.warning(self, cfg.programName, "The ethogram is not set!")
        return

    if self.pj[cfg.OBSERVATIONS][self.observationId][cfg.TYPE] == cfg.IMAGES:
        current_time = self.image_idx + 1
    else:
        current_time = self.getLaps()

    editWindow = DlgEditEvent(
        observation_type=self.pj[cfg.OBSERVATIONS][self.observationId][cfg.TYPE],
        time_value=dec("NaN"),
        image_idx=0,
        current_time=current_time,
        time_format=self.timeFormat,
        show_set_current_time=True,
    )
    editWindow.setWindowTitle("Add a new event")

    sortedSubjects = [cfg.NO_FOCAL_SUBJECT] + sorted([self.pj[cfg.SUBJECTS][x][cfg.SUBJECT_NAME] for x in self.pj[cfg.SUBJECTS]])

    editWindow.cobSubject.addItems(sortedSubjects)
    if self.currentSubject:
        editWindow.cobSubject.setCurrentIndex(editWindow.cobSubject.findText(self.currentSubject, Qt.MatchFixedString))

    sortedCodes = sorted([self.pj[cfg.ETHOGRAM][x][cfg.BEHAVIOR_CODE] for x in self.pj[cfg.ETHOGRAM]])

    editWindow.cobCode.addItems(sortedCodes)

    if editWindow.exec_():  # button OK
        # MEDIA / LIVE
        if self.pj[cfg.OBSERVATIONS][self.observationId][cfg.TYPE] in (cfg.MEDIA, cfg.LIVE):
            newTime = editWindow.time_widget.get_time()
            if newTime is None:
                QMessageBox.warning(
                    self,
                    cfg.programName,
                    ("Select a time format"),
                )
                return

            for idx in self.pj[cfg.ETHOGRAM]:
                if self.pj[cfg.ETHOGRAM][idx][cfg.BEHAVIOR_CODE] == editWindow.cobCode.currentText():
                    event = self.full_event(idx)

                    event[cfg.SUBJECT] = (
                        "" if editWindow.cobSubject.currentText() == cfg.NO_FOCAL_SUBJECT else editWindow.cobSubject.currentText()
                    )
                    if editWindow.leComment.toPlainText():
                        event[cfg.COMMENT] = editWindow.leComment.toPlainText()

                    # determine the frame index
                    if self.playerType == cfg.MEDIA:
                        mem_time = self.getLaps()
                        if not self.seek_mediaplayer(newTime):
                            time.sleep(0.1)
                            frame_idx = self.get_frame_index()
                            event[cfg.FRAME_INDEX] = frame_idx
                            self.seek_mediaplayer(mem_time)

                    write_event.write_event(self, event, newTime)
                    break

            self.update_realtime_plot(force_plot=True)
            """
            if hasattr(self, "plot_events"):
                if not self.plot_events.visibleRegion().isEmpty():
                    self.plot_events.events_list = self.pj[cfg.OBSERVATIONS][self.observationId][cfg.EVENTS]
                    self.plot_events.plot_events(float(self.getLaps()))
            """

            """
            # update subjects table
            self.currentStates = util.get_current_states_modifiers_by_subject(
                util.state_behavior_codes(self.pj[cfg.ETHOGRAM]),
                self.pj[cfg.OBSERVATIONS][self.observationId][cfg.EVENTS],
                dict(self.pj[cfg.SUBJECTS], **{"": {"name": ""}}),  # add no focal subject
                newTime,
                include_modifiers=True,
            )
            subject_idx = self.subject_name_index[self.currentSubject] if self.currentSubject else ""
            self.lbCurrentStates.setText(", ".join(self.currentStates[subject_idx]))
            self.show_current_states_in_subjects_table()
            """

        # IMAGES
        if self.pj[cfg.OBSERVATIONS][self.observationId][cfg.TYPE] == cfg.IMAGES:
            new_index = editWindow.sb_image_idx.value()
            if new_index == 0:
                QMessageBox.warning(self, cfg.programName, "The image index cannot be null")
                return

            for idx in self.pj[cfg.ETHOGRAM]:
                if self.pj[cfg.ETHOGRAM][idx][cfg.BEHAVIOR_CODE] == editWindow.cobCode.currentText():
                    event = self.full_event(idx)

                    event[cfg.SUBJECT] = (
                        "" if editWindow.cobSubject.currentText() == cfg.NO_FOCAL_SUBJECT else editWindow.cobSubject.currentText()
                    )
                    if editWindow.leComment.toPlainText():
                        event[cfg.COMMENT] = editWindow.leComment.toPlainText()

                    if self.playerType != cfg.VIEWER_IMAGES:
                        event[cfg.IMAGE_PATH] = self.images_list[new_index]
                    else:
                        event[cfg.IMAGE_PATH] = ""

                    event[cfg.IMAGE_INDEX] = new_index

                    time_ = dec("NaN")
                    if self.pj[cfg.OBSERVATIONS][self.observationId].get(cfg.USE_EXIF_DATE, False):
                        if self.playerType != cfg.VIEWER_IMAGES:
                            exif_date_time = util.extract_exif_DateTimeOriginal(self.images_list[new_index])
                            if exif_date_time != -1:
                                time_ = exif_date_time

                                # check if first value must be substracted
                                if self.pj[cfg.OBSERVATIONS][self.observationId].get(cfg.SUBSTRACT_FIRST_EXIF_DATE, True):
                                    time_ -= self.image_time_ref

                    elif self.pj[cfg.OBSERVATIONS][self.observationId].get(cfg.TIME_LAPSE, 0):
                        time_ = new_index * self.pj[cfg.OBSERVATIONS][self.observationId].get(cfg.TIME_LAPSE, 0)

                    write_event.write_event(self, event, dec(time_).quantize(dec("0.001"), rounding=ROUND_DOWN))

                    break

    if self.pause_before_addevent:
        # restart media
        if self.pj[cfg.OBSERVATIONS][self.observationId][cfg.TYPE] == cfg.MEDIA and self.playerType == cfg.MEDIA:
            if memState:
                self.play_video()


def find_events(self):
    """
    find in events
    """

    self.find_dialog = dialog.FindInEvents()
    # list of rows to find
    self.find_dialog.rowsToFind = set([self.tv_idx2events_idx[item.row()] for item in self.tv_events.selectedIndexes()])
    self.find_dialog.currentIdx = -1
    self.find_dialog.clickSignal.connect(self.click_signal_find_in_events)
    self.find_dialog.setWindowFlags(Qt.WindowStaysOnTopHint)
    self.find_dialog.show()


def find_replace_events(self):
    """
    find and replace in events
    """
    fill_events_undo_list(self, "Undo Find/Replace operations")
    self.find_replace_dialog = dialog.FindReplaceEvents()
    self.find_replace_dialog.currentIdx = -1
    self.find_replace_dialog.currentIdx_idx = -1
    # list of rows to find/replace
    self.find_replace_dialog.rowsToFind = set([self.tv_idx2events_idx[item.row()] for item in self.tv_events.selectedIndexes()])
    self.find_replace_dialog.clickSignal.connect(self.click_signal_find_replace_in_events)
    self.find_replace_dialog.setWindowFlags(Qt.WindowStaysOnTopHint)
    self.find_replace_dialog.show()


def filter_events(self):
    """
    filter coded events and subjects
    """

    parameters = select_subj_behav.choose_obs_subj_behav_category(
        self,
        selected_observations=[],  # empty selection of observations for selecting all subjects and behaviors
        show_include_modifiers=False,
        show_exclude_non_coded_behaviors=False,
        by_category=False,
    )
    if parameters == {}:
        return

    self.filtered_subjects = parameters[cfg.SELECTED_SUBJECTS][:]
    if cfg.NO_FOCAL_SUBJECT in self.filtered_subjects:
        self.filtered_subjects.append("")
    self.filtered_behaviors = parameters[cfg.SELECTED_BEHAVIORS][:]

    logging.debug(f"self.filtered_behaviors: {self.filtered_behaviors}")

    self.load_tw_events(self.observationId)
    self.dwEvents.setWindowTitle(f"Events for “{self.observationId}” observation (filtered)")


def show_all_events(self) -> None:
    """
    show all events (disable filter)
    """
    self.filtered_subjects = []
    self.filtered_behaviors = []
    self.load_tw_events(self.observationId)
    self.dwEvents.setWindowTitle(f"Events for “{self.observationId}” observation")


def fill_events_undo_list(self, operation_description: str) -> None:
    """
    fill the undo events list for Undo function (CTRL + Z)
    """
    logging.debug("fill_events_undo_list function")

    self.undo_queue.append(copy.deepcopy(self.pj[cfg.OBSERVATIONS][self.observationId][cfg.EVENTS]))

    self.undo_description.append(operation_description)

    self.actionUndo.setText(operation_description)
    self.actionUndo.setEnabled(True)

    logging.debug(f"{operation_description} added to undo events list")

    if len(self.undo_queue) > cfg.MAX_UNDO_QUEUE:
        self.undo_queue.popleft()
        self.undo_description.popleft()
        logging.debug("Max events undo ")


def undo_event_operation(self) -> None:
    """
    undo operation on event(s)
    """

    logging.debug("Undo event operation function")

    if len(self.undo_queue) == 0:
        self.statusbar.showMessage("The Undo buffer is empty", 5000)
        return

    events = self.undo_queue.pop()

    operation_description = self.undo_description.pop()
    self.pj[cfg.OBSERVATIONS][self.observationId][cfg.EVENTS] = events[:]
    self.project_changed()

    self.statusbar.showMessage(operation_description, 5000)

    logging.debug(operation_description)

    # reload all events in tw
    self.load_tw_events(self.observationId)

    self.update_realtime_plot(force_plot=True)

    if not len(self.undo_queue):
        self.actionUndo.setText("Undo")
        self.actionUndo.setEnabled(False)
    else:
        self.actionUndo.setText(self.undo_description[-1])


def delete_all_events(self):
    """
    delete all (filtered) events in current observation
    """

    if not self.observationId:
        self.no_observation()
        return

    if not self.tv_idx2events_idx:
        QMessageBox.warning(self, cfg.programName, "No events to delete")
        return

    if (
        dialog.MessageDialog(
            cfg.programName,
            ("Confirm the deletion of all (filtered) events in the current observation?<br>Filters do not apply!"),
            [cfg.YES, cfg.NO],
        )
        == cfg.YES
    ):
        # fill the undo list
        fill_events_undo_list(self, "Undo 'Delete all events'")

        self.pj[cfg.OBSERVATIONS][self.observationId][cfg.EVENTS] = [
            event
            for event_idx, event in enumerate(self.pj[cfg.OBSERVATIONS][self.observationId][cfg.EVENTS])
            if event_idx not in self.tv_idx2events_idx
        ]

        self.update_realtime_plot(force_plot=True)

        self.project_changed()
        self.load_tw_events(self.observationId)


def delete_selected_events(self):
    """
    delete selected events
    """

    if not self.observationId:
        self.no_observation()
        return

    logging.debug("begin function delete_selected_events")

    if not self.tv_events.selectedIndexes():
        QMessageBox.warning(self, cfg.programName, "No event selected!")
    else:
        # list of rows to delete (set for unique)
        # fill the undo list
        fill_events_undo_list(self, "Undo 'Delete selected events'")

        rows_to_delete: list = []
        for row in set([item.row() for item in self.tv_events.selectedIndexes()]):
            rows_to_delete.append(self.tv_idx2events_idx[row])

        self.pj[cfg.OBSERVATIONS][self.observationId][cfg.EVENTS] = [
            event
            for event_idx, event in enumerate(self.pj[cfg.OBSERVATIONS][self.observationId][cfg.EVENTS])
            if event_idx not in rows_to_delete
        ]

        self.update_realtime_plot(force_plot=True)

        self.project_changed()
        self.load_tw_events(self.observationId)


def select_events_between_activated(self):
    """
    select events between a time interval
    """

    def parseTime(txt):
        """
        parse time in string (should be 00:00:00.000 or in seconds)
        """
        if ":" in txt:
            qtime = QTime.fromString(txt, "hh:mm:ss.zzz")

            if qtime.toString():
                timeSeconds = util.time2seconds(qtime.toString("hh:mm:ss.zzz"))
            else:
                return None
        else:
            try:
                timeSeconds = dec(txt)
            except InvalidOperation:
                return None
        return timeSeconds

    if not self.tv_idx2events_idx:
        QMessageBox.warning(self, cfg.programName, "There are no events to select")
        return

    text, ok = QInputDialog.getText(
        self,
        "Select events in time interval",
        "Interval: (example: 12.5-14.7 or 02:45.780-03:15.120)",
        QLineEdit.Normal,
        "",
    )

    if ok and text != "":
        if "-" not in text:
            QMessageBox.critical(self, cfg.programName, "Use minus sign (-) to separate initial value from final value")
            return

        while " " in text:
            text = text.replace(" ", "")

        from_, to_ = text.split("-")[0:2]
        from_sec = parseTime(from_)
        if not from_sec:
            QMessageBox.critical(self, cfg.programName, f"Time value not recognized: {from_}")
            return
        to_sec = parseTime(to_)
        if not to_sec:
            QMessageBox.critical(self, cfg.programName, f"Time value not recognized: {to_}")
            return
        if to_sec < from_sec:
            QMessageBox.critical(self, cfg.programName, "The initial time is greater than the final time")
            return

        self.tv_events.clearSelection()
        self.tv_events.setSelectionMode(QAbstractItemView.MultiSelection)

        # for r in range(self.tv_events.rowCount()):
        # for idx, event in enumerate(self.pj[cfg.OBSERVATIONS][self.observationId][cfg.EVENTS]):
        for tv_idx in range(len(self.tv_idx2events_idx)):
            time = self.pj[cfg.OBSERVATIONS][self.observationId][cfg.EVENTS][self.tv_idx2events_idx[tv_idx]][
                cfg.PJ_OBS_FIELDS[self.playerType][cfg.TIME]
            ]

            if from_sec <= time <= to_sec:
                self.tv_events.selectRow(tv_idx)

        self.tv_events.setSelectionMode(QAbstractItemView.ExtendedSelection)


def edit_selected_events(self):
    """
    edit one or more selected events for subject, behavior and/or comment
    """
    # list of rows to edit

    tvevents_rows_to_edit = set([index.row() for index in self.tv_events.selectionModel().selectedIndexes()])

    if not len(tvevents_rows_to_edit):
        QMessageBox.warning(self, cfg.programName, "No event selected!")

    elif len(tvevents_rows_to_edit) == 1:  # 1 event selected
        edit_event(self)

    else:  # editing of more events
        dialog_window = EditSelectedEvents()
        dialog_window.all_behaviors = sorted([self.pj[cfg.ETHOGRAM][x][cfg.BEHAVIOR_CODE] for x in self.pj[cfg.ETHOGRAM]])

        dialog_window.all_subjects = [cfg.NO_FOCAL_SUBJECT] + [
            self.pj[cfg.SUBJECTS][str(k)][cfg.SUBJECT_NAME] for k in sorted([int(x) for x in self.pj[cfg.SUBJECTS].keys()])
        ]

        if dialog_window.exec_():
            # fill the undo list
            fill_events_undo_list(self, "Undo 'Edit selected event(s)'")

            tsb_to_edit: list = []
            for row in tvevents_rows_to_edit:
                tsb_to_edit.append(self.tv_idx2events_idx[row])

            behavior_codes: list = []
            modifiers_mem: list = []
            mem_event_idx: list = []
            for idx, event in enumerate(self.pj[cfg.OBSERVATIONS][self.observationId][cfg.EVENTS]):
                if idx in tsb_to_edit:
                    new_event = list(event)
                    if dialog_window.rbSubject.isChecked():
                        if dialog_window.newText.selectedItems()[0].text() == cfg.NO_FOCAL_SUBJECT:
                            new_event[cfg.EVENT_SUBJECT_FIELD_IDX] = ""
                        else:
                            new_event[cfg.EVENT_SUBJECT_FIELD_IDX] = dialog_window.newText.selectedItems()[0].text()
                    if dialog_window.rbBehavior.isChecked():
                        new_event[cfg.EVENT_BEHAVIOR_FIELD_IDX] = dialog_window.newText.selectedItems()[0].text()
                    if dialog_window.rbComment.isChecked():
                        new_event[cfg.EVENT_COMMENT_FIELD_IDX] = dialog_window.commentText.text()

                    if new_event[cfg.EVENT_BEHAVIOR_FIELD_IDX] not in behavior_codes:
                        behavior_codes.append(new_event[cfg.EVENT_BEHAVIOR_FIELD_IDX])

                    if new_event[cfg.EVENT_MODIFIER_FIELD_IDX] not in modifiers_mem:
                        modifiers_mem.append(new_event[cfg.EVENT_MODIFIER_FIELD_IDX])

                    self.pj[cfg.OBSERVATIONS][self.observationId][cfg.EVENTS][idx] = list(new_event)
                    mem_event_idx.append(idx)
                    self.project_changed()

            # check if behavior is unique for editing modifiers
            if len(behavior_codes) == 1:
                # get behavior index
                for idx in self.pj[cfg.ETHOGRAM]:
                    if self.pj[cfg.ETHOGRAM][idx][cfg.BEHAVIOR_CODE] == behavior_codes[0]:
                        break
                else:
                    return

                event = self.full_event(idx)

                if len(modifiers_mem) == 1:
                    current_modifier = modifiers_mem[0]
                else:
                    current_modifier = ""

                if event["modifiers"]:
                    modifiers_selector = select_modifiers.ModifiersList(behavior_codes[0], eval(str(event["modifiers"])), current_modifier)

                    r = modifiers_selector.exec_()
                    if r:
                        selected_modifiers = modifiers_selector.get_modifiers()

                        modifier_str = ""
                        for idx1 in util.sorted_keys(selected_modifiers):
                            if modifier_str:
                                modifier_str += "|"
                            if selected_modifiers[idx1]["type"] in (cfg.SINGLE_SELECTION, cfg.MULTI_SELECTION):
                                modifier_str += ",".join(selected_modifiers[idx1].get("selected", ""))
                            if selected_modifiers[idx1]["type"] in (cfg.NUMERIC_MODIFIER, cfg.EXTERNAL_DATA_MODIFIER):
                                modifier_str += selected_modifiers[idx1].get("selected", "NA")

                else:  # delete current modifier(s)
                    modifier_str = ""

                # set new modifier
                for idx, event in enumerate(self.pj[cfg.OBSERVATIONS][self.observationId][cfg.EVENTS]):
                    if idx in mem_event_idx:
                        self.pj[cfg.OBSERVATIONS][self.observationId][cfg.EVENTS][idx][cfg.EVENT_MODIFIER_FIELD_IDX] = modifier_str

            self.load_tw_events(self.observationId)

            self.update_realtime_plot(force_plot=True)


def edit_event(self):
    """
    edit event corresponding to the selected row in tv_events
    """

    if not self.observationId:
        self.no_observation()
        return

    if not self.tv_events.selectionModel().selectedIndexes():
        QMessageBox.warning(self, cfg.programName, "Select an event to edit")
        return

    if self.pause_before_addevent:
        # pause media
        if self.pj[cfg.OBSERVATIONS][self.observationId][cfg.TYPE] == cfg.MEDIA and self.playerType == cfg.MEDIA:
            player_mem_state = self.is_playing()
            if player_mem_state:
                self.pause_video()

    tvevents_row = self.tv_events.selectionModel().selectedIndexes()[0].row()

    pj_event_idx = self.tv_idx2events_idx[tvevents_row]

    time_value = self.pj[cfg.OBSERVATIONS][self.observationId][cfg.EVENTS][pj_event_idx][
        cfg.PJ_OBS_FIELDS[self.pj[cfg.OBSERVATIONS][self.observationId][cfg.TYPE]][cfg.TIME]
    ]

    image_idx = None
    if self.pj[cfg.OBSERVATIONS][self.observationId][cfg.TYPE] in (cfg.IMAGES):
        image_idx = self.pj[cfg.OBSERVATIONS][self.observationId][cfg.EVENTS][pj_event_idx][
            cfg.PJ_OBS_FIELDS[self.pj[cfg.OBSERVATIONS][self.observationId][cfg.TYPE]][cfg.IMAGE_INDEX]
        ]

    if self.pj[cfg.OBSERVATIONS][self.observationId][cfg.TYPE] == cfg.IMAGES:
        current_value = self.image_idx + 1
    else:
        current_value = self.getLaps()

    # get exif date time
    if self.pj[cfg.OBSERVATIONS][self.observationId].get(cfg.USE_EXIF_DATE, False):
        exif_date_time = util.extract_exif_DateTimeOriginal(self.images_list[self.image_idx])
    else:
        exif_date_time = None

    edit_window = DlgEditEvent(
        observation_type=self.pj[cfg.OBSERVATIONS][self.observationId][cfg.TYPE],
        time_value=time_value,
        image_idx=image_idx,
        current_time=current_value,
        time_format=self.timeFormat,
        show_set_current_time=True,
        exif_date_time=exif_date_time,
    )
    edit_window.setWindowTitle("Edit event")

    # time
    if time_value.is_nan():
        edit_window.cb_set_time_na.setChecked(True)

    # remove visibility of 'set current time' widget if VIEWER mode
    if self.playerType in (cfg.VIEWER_MEDIA, cfg.VIEWER_LIVE, cfg.VIEWER_IMAGES):
        edit_window.pb_set_to_current_time.setVisible(False)

    # subjects
    sorted_subjects = [cfg.NO_FOCAL_SUBJECT] + sorted([self.pj[cfg.SUBJECTS][x][cfg.SUBJECT_NAME] for x in self.pj[cfg.SUBJECTS]])
    edit_window.cobSubject.addItems(sorted_subjects)
    if self.pj[cfg.OBSERVATIONS][self.observationId][cfg.EVENTS][pj_event_idx][cfg.EVENT_SUBJECT_FIELD_IDX] == "":  # no focal subject
        edit_window.cobSubject.setCurrentIndex(sorted_subjects.index(cfg.NO_FOCAL_SUBJECT))
    elif self.pj[cfg.OBSERVATIONS][self.observationId][cfg.EVENTS][pj_event_idx][cfg.EVENT_SUBJECT_FIELD_IDX] in sorted_subjects:
        edit_window.cobSubject.setCurrentIndex(
            sorted_subjects.index(self.pj[cfg.OBSERVATIONS][self.observationId][cfg.EVENTS][pj_event_idx][cfg.EVENT_SUBJECT_FIELD_IDX])
        )
    else:
        QMessageBox.warning(
            self,
            cfg.programName,
            (
                "The subject "
                f"<b>{self.pj[cfg.OBSERVATIONS][self.observationId][cfg.EVENTS][pj_event_idx][cfg.EVENT_SUBJECT_FIELD_IDX]}</b> "
                "does not exist more in the subject's list"
            ),
        )
        edit_window.cobSubject.setCurrentIndex(0)

    # behaviors
    sortedCodes = sorted([self.pj[cfg.ETHOGRAM][x][cfg.BEHAVIOR_CODE] for x in self.pj[cfg.ETHOGRAM]])
    edit_window.cobCode.addItems(sortedCodes)
    # check if selected code is in code's list (no modification of codes)
    if self.pj[cfg.OBSERVATIONS][self.observationId][cfg.EVENTS][pj_event_idx][cfg.EVENT_BEHAVIOR_FIELD_IDX] in sortedCodes:
        edit_window.cobCode.setCurrentIndex(
            sortedCodes.index(self.pj[cfg.OBSERVATIONS][self.observationId][cfg.EVENTS][pj_event_idx][cfg.EVENT_BEHAVIOR_FIELD_IDX])
        )
    else:
        msg: str = (
            f"The behaviour {self.pj[cfg.OBSERVATIONS][self.observationId][cfg.EVENTS][pj_event_idx][cfg.EVENT_BEHAVIOR_FIELD_IDX]} "
            "does not exist longer in the ethogram"
        )
        logging.warning(msg)

        QMessageBox.warning(
            self,
            cfg.programName,
            msg,
        )
        edit_window.cobCode.setCurrentIndex(0)

    logging.debug(
        f"original modifiers: {self.pj[cfg.OBSERVATIONS][self.observationId][cfg.EVENTS][pj_event_idx][cfg.EVENT_MODIFIER_FIELD_IDX]}"
    )

    # # frame index
    # frame_idx = read_event_field(
    #     self.pj[cfg.OBSERVATIONS][self.observationId][cfg.EVENTS][pj_event_idx],
    #     self.pj[cfg.OBSERVATIONS][self.observationId][cfg.TYPE],
    #     cfg.FRAME_INDEX,
    # )
    # edit_window.sb_frame_idx.setValue(0 if frame_idx in (cfg.NA, None) else frame_idx)
    # if frame_idx in (cfg.NA, None):
    #     edit_window.cb_set_frame_idx_na.setChecked(True)

    # comment
    edit_window.leComment.setPlainText(self.pj[cfg.OBSERVATIONS][self.observationId][cfg.EVENTS][pj_event_idx][cfg.EVENT_COMMENT_FIELD_IDX])

    flag_ok = False  # for looping until event is OK or Cancel pressed
    while True:
        if edit_window.exec():  # button OK
            self.project_changed()

            new_time = edit_window.time_widget.get_time()

            if edit_window.cb_set_time_na.isChecked():
                new_time = dec("NaN")

            # MEDIA / LIVE
            if self.pj[cfg.OBSERVATIONS][self.observationId][cfg.TYPE] in (cfg.MEDIA, cfg.LIVE):
                if new_time.is_nan():
                    QMessageBox.warning(
                        self,
                        cfg.programName,
                        ("Select a time format"),
                    )
                    continue

                for key in self.pj[cfg.ETHOGRAM]:
                    if self.pj[cfg.ETHOGRAM][key][cfg.BEHAVIOR_CODE] == edit_window.cobCode.currentText():
                        event = self.full_event(key)
                        # subject
                        event[cfg.SUBJECT] = (
                            "" if edit_window.cobSubject.currentText() == cfg.NO_FOCAL_SUBJECT else edit_window.cobSubject.currentText()
                        )

                        event[cfg.COMMENT] = edit_window.leComment.toPlainText()

                        # determine the new frame index
                        if self.pj[cfg.OBSERVATIONS][self.observationId][cfg.TYPE] == cfg.MEDIA:
                            if self.playerType == cfg.MEDIA:
                                mem_time = self.getLaps()
                                if not self.seek_mediaplayer(new_time):
                                    time.sleep(0.1)
                                    frame_idx = self.get_frame_index()
                                    event[cfg.FRAME_INDEX] = frame_idx
                                    self.seek_mediaplayer(mem_time)

                            # if not edit_window.sb_frame_idx.value() or edit_window.cb_set_frame_idx_na.isChecked():
                            #     event[cfg.FRAME_INDEX] = cfg.NA
                            # else:
                            #     if self.playerType == cfg.MEDIA:
                            #         mem_time = self.getLaps()
                            #         if not self.seek_mediaplayer(new_time):
                            #             frame_idx = self.get_frame_index()
                            #             event[cfg.FRAME_INDEX] = frame_idx
                            #             self.seek_mediaplayer(mem_time)
                            #
                            #     # event[cfg.FRAME_INDEX] = edit_window.sb_frame_idx.value()

                        event["row"] = pj_event_idx
                        event["original_modifiers"] = self.pj[cfg.OBSERVATIONS][self.observationId][cfg.EVENTS][pj_event_idx][
                            cfg.PJ_OBS_FIELDS[self.playerType][cfg.MODIFIER]
                        ]

                        r = write_event.write_event(self, event, new_time)

                        # scroll tv events
                        index = self.tv_events.model().index(pj_event_idx, 0)
                        self.tv_events.scrollTo(index, QAbstractItemView.EnsureVisible)

                        if r == 1:  # same event already present
                            continue
                        if not r:
                            flag_ok = True
                        break

                self.update_realtime_plot(force_plot=True)

            # IMAGES
            if self.pj[cfg.OBSERVATIONS][self.observationId][cfg.TYPE] == cfg.IMAGES:
                new_index = edit_window.sb_image_idx.value()

                for key in self.pj[cfg.ETHOGRAM]:
                    if self.pj[cfg.ETHOGRAM][key][cfg.BEHAVIOR_CODE] == edit_window.cobCode.currentText():
                        event = self.full_event(key)
                        event[cfg.TIME] = new_time

                        event[cfg.SUBJECT] = (
                            "" if edit_window.cobSubject.currentText() == cfg.NO_FOCAL_SUBJECT else edit_window.cobSubject.currentText()
                        )
                        event[cfg.COMMENT] = edit_window.leComment.toPlainText()
                        event["row"] = pj_event_idx
                        event["original_modifiers"] = self.pj[cfg.OBSERVATIONS][self.observationId][cfg.EVENTS][pj_event_idx][
                            cfg.PJ_OBS_FIELDS[self.playerType][cfg.MODIFIER]
                        ]

                        # not editable yet. Read previous value
                        event[cfg.IMAGE_PATH] = read_event_field(
                            self.pj[cfg.OBSERVATIONS][self.observationId][cfg.EVENTS][pj_event_idx],
                            self.pj[cfg.OBSERVATIONS][self.observationId][cfg.TYPE],
                            cfg.IMAGE_PATH,
                        )

                        event[cfg.IMAGE_INDEX] = new_index

                        r = write_event.write_event(self, event, event[cfg.TIME].quantize(dec("0.001"), rounding=ROUND_DOWN))
                        if r == 1:  # same event already present
                            continue
                        if not r:
                            flag_ok = True
                        break

        else:  # Cancel button
            flag_ok = True

        if flag_ok:
            break

    if self.pause_before_addevent:
        # restart media
        if self.pj[cfg.OBSERVATIONS][self.observationId][cfg.TYPE] == cfg.MEDIA and self.playerType == cfg.MEDIA:
            if player_mem_state:
                self.play_video()


def edit_time_selected_events(self):
    """
    shift time of one or more selected events
    """

    tvevents_rows_to_shift = set([index.row() for index in self.tv_events.selectionModel().selectedIndexes()])
    if not len(tvevents_rows_to_shift):
        QMessageBox.warning(self, cfg.programName, "No event selected!")
        return

    w = dialog.Ask_time(0)
    w.setWindowTitle("Shift time of selected event(s)")
    w.label.setText("Amount of time to add or substract")

    if not w.exec_():
        return

    d = w.time_widget.get_time()
    if d.is_nan() or not d:
        return

    if ":" in util.smart_time_format(abs(d)):
        smart_d = f"{util.smart_time_format(abs(d))}"
    else:
        smart_d = f"{d} seconds"

    if (
        dialog.MessageDialog(
            cfg.programName,
            (f"Confirm the {'addition' if d > 0 else 'subtraction'} of {smart_d} to all selected events in the current observation?"),
            [cfg.YES, cfg.NO],
        )
        == cfg.NO
    ):
        return

    # fill the undo list
    fill_events_undo_list(self, "Undo 'Edit time'")

    mem_time = self.getLaps()
    for tw_event_idx in tvevents_rows_to_shift:
        pj_event_idx = self.tv_idx2events_idx[tw_event_idx]

        self.pj[cfg.OBSERVATIONS][self.observationId][cfg.EVENTS][pj_event_idx][cfg.PJ_OBS_FIELDS[self.playerType][cfg.TIME]] += dec(
            f"{d:.3f}"
        )
        # set new frame index
        if self.playerType == cfg.MEDIA:
            if not self.seek_mediaplayer(
                self.pj[cfg.OBSERVATIONS][self.observationId][cfg.EVENTS][pj_event_idx][cfg.PJ_OBS_FIELDS[self.playerType][cfg.TIME]]
            ):
                # determine the new frame index
                time.sleep(0.1)
                frame_idx = self.get_frame_index()
                if len(self.pj[cfg.OBSERVATIONS][self.observationId][cfg.EVENTS][pj_event_idx]) == 6:
                    self.pj[cfg.OBSERVATIONS][self.observationId][cfg.EVENTS][pj_event_idx][-1] = frame_idx
                elif len(self.pj[cfg.OBSERVATIONS][self.observationId][cfg.EVENTS][pj_event_idx]) == 5:
                    self.pj[cfg.OBSERVATIONS][self.observationId][cfg.EVENTS][pj_event_idx].append(frame_idx)

        self.project_changed()

    if self.playerType == cfg.MEDIA:
        self.seek_mediaplayer(mem_time)

    self.pj[cfg.OBSERVATIONS][self.observationId][cfg.EVENTS] = sorted(self.pj[cfg.OBSERVATIONS][self.observationId][cfg.EVENTS])
    self.load_tw_events(self.observationId)

    self.update_realtime_plot(force_plot=True)


def copy_selected_events(self):
    """
    copy selected events from project to clipboard
    """

    logging.debug("Copy selected events to clipboard")

    tvevents_rows_to_copy = set([index.row() for index in self.tv_events.selectionModel().selectedIndexes()])
    if not len(tvevents_rows_to_copy):
        QMessageBox.warning(self, cfg.programName, "No event selected!")
        return

    pj_event_idx_to_copy: list = [self.tv_idx2events_idx[row] for row in tvevents_rows_to_copy]

    copied_events: list = []
    for idx, event in enumerate(self.pj[cfg.OBSERVATIONS][self.observationId][cfg.EVENTS]):
        if idx in pj_event_idx_to_copy:
            if self.playerType in (cfg.MEDIA, cfg.VIEWER_MEDIA) and len(event) < len(cfg.MEDIA_PJ_EVENTS_FIELDS):
                copied_events.append("\t".join([str(x) for x in event + [cfg.NA]]))
            else:
                copied_events.append("\t".join([str(x) for x in event]))

    cb = QApplication.clipboard()
    cb.clear(mode=QClipboard.Mode.Clipboard)
    cb.setText("\n".join(copied_events), mode=QClipboard.Mode.Clipboard)

    logging.debug("Selected events copied in clipboard")


def paste_clipboard_to_events(self):
    """
    paste clipboard to events
    """

    cb = QApplication.clipboard()
    cb_text = cb.text()
    cb_text_splitted = cb_text.split("\n")
    length: list = []
    content: list = []
    for line in cb_text_splitted:
        length.append(len(line.split("\t")))
        content.append(line.split("\t"))

    if set(length) != set([len(cfg.PJ_EVENTS_FIELDS[self.playerType])]):
        msg_box = QMessageBox(
            QMessageBox.Warning,
            cfg.programName,
            (
                "The clipboard does not contain events!<br>"
                f"For an observation from <b>{self.playerType}</b> "
                f"the events must be organized in {len(cfg.PJ_EVENTS_FIELDS[self.playerType])} columns separated by <TAB> character"
            ),
        )
        msg_box.setWindowFlags(msg_box.windowFlags() | Qt.WindowStaysOnTopHint)
        msg_box.exec()

        return

    for event in content:
        # convert time in decimal
        event[cfg.EVENT_TIME_FIELD_IDX] = dec(event[cfg.EVENT_TIME_FIELD_IDX])
        for idx, _ in enumerate(event):
            if cfg.PJ_EVENTS_FIELDS[self.playerType][idx] in (cfg.FRAME_INDEX, cfg.IMAGE_INDEX):
                try:
                    event[idx] = int(event[idx])
                except ValueError:
                    pass

        # skip if event already present
        if event in self.pj[cfg.OBSERVATIONS][self.observationId][cfg.EVENTS]:
            continue

        self.pj[cfg.OBSERVATIONS][self.observationId][cfg.EVENTS].append(event)

        self.project_changed()

    self.pj[cfg.OBSERVATIONS][self.observationId][cfg.EVENTS] = sorted(self.pj[cfg.OBSERVATIONS][self.observationId][cfg.EVENTS])
    self.load_tw_events(self.observationId)

    self.update_realtime_plot(force_plot=True)


def read_event_field(event: list, player_type: str, field_type: str) -> Union[str, None, int, dec]:
    """
    return value of field for event or NA if not available
    """
    if field_type not in cfg.PJ_EVENTS_FIELDS[player_type]:
        return None
    if cfg.PJ_OBS_FIELDS[player_type][field_type] < len(event):
        return event[cfg.PJ_OBS_FIELDS[player_type][field_type]]
    else:
        return cfg.NA


def add_frame_indexes(self):
    """
    add frame indexes for all events
    """
    if self.pj[cfg.OBSERVATIONS][self.observationId][cfg.TYPE] != cfg.MEDIA:
        return
    if self.playerType != cfg.MEDIA:
        return

    mem_time = self.getLaps()
    for idx, event in enumerate(self.pj[cfg.OBSERVATIONS][self.observationId][cfg.EVENTS]):
        if event[0] == "NA":
            continue
        if not self.seek_mediaplayer(event[0]):
            time.sleep(0.1)
            self.pj[cfg.OBSERVATIONS][self.observationId][cfg.EVENTS][idx][cfg.PJ_OBS_FIELDS[cfg.MEDIA][cfg.FRAME_INDEX]] = (
                self.get_frame_index()
            )

    self.seek_mediaplayer(mem_time)

    self.load_tw_events(self.observationId)



"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2025 Olivier Friard


  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
  MA 02110-1301, USA.

"""

import logging
from typing import Tuple

from PySide6.QtCore import Qt
from PySide6.QtWidgets import QAbstractItemView

from . import config as cfg
from . import gui_utilities, observations_list, project_functions
from . import utilities as util


def select_observations2(self, mode: str, windows_title: str = "") -> Tuple[str, list]:
    """
    allow user to select observations
    mode: accepted values: OPEN, EDIT, SINGLE, MULTIPLE, SELECT1

    Args:
        pj (dict): BORIS project dictionary
        mode (str): mode for selection: OPEN, EDIT, SINGLE, MULTIPLE, SELECT1
        windows_title (str): title for windows

    Returns:
        str: selected mode: OPEN, EDIT, VIEW
        list: list of selected observations
    """

    pj = self.pj

    fields_list = ["id", "date", "description", "subjects", "observation duration", "exhaustivity %", "media"]
    indep_var_header, column_type = [], [cfg.TEXT, cfg.TEXT, cfg.TEXT, cfg.TEXT, cfg.NUMERIC, cfg.NUMERIC, cfg.TEXT]

    if cfg.INDEPENDENT_VARIABLES in pj:
        for idx in util.sorted_keys(pj[cfg.INDEPENDENT_VARIABLES]):
            indep_var_header.append(pj[cfg.INDEPENDENT_VARIABLES][idx]["label"])
            column_type.append(pj[cfg.INDEPENDENT_VARIABLES][idx]["type"])

    state_events_list = util.state_behavior_codes(pj[cfg.ETHOGRAM])

    data: list = []
    not_paired: list = []

    # check if observations changed
    if hash(str(self.pj[cfg.OBSERVATIONS])) != self.mem_hash_obs:
        logging.debug("observations changed")

        for obs in sorted(list(pj[cfg.OBSERVATIONS].keys())):
            date = pj[cfg.OBSERVATIONS][obs]["date"].replace("T", " ")
            descr = util.eol2space(pj[cfg.OBSERVATIONS][obs][cfg.DESCRIPTION])

            # subjects
            observed_subjects = [cfg.NO_FOCAL_SUBJECT if x == "" else x for x in project_functions.extract_observed_subjects(pj, [obs])]

            subjectsList = ", ".join(observed_subjects)

            # observed time interval
            interval = project_functions.observed_interval(pj[cfg.OBSERVATIONS][obs])
            observed_interval_str = str(round(interval[1] - interval[0], 3))

            # media
            media: str = ""
            if pj[cfg.OBSERVATIONS][obs][cfg.TYPE] == cfg.MEDIA:
                media_list: list = []
                if pj[cfg.OBSERVATIONS][obs][cfg.FILE]:
                    for player in sorted(pj[cfg.OBSERVATIONS][obs][cfg.FILE].keys()):
                        for media in pj[cfg.OBSERVATIONS][obs][cfg.FILE][player]:
                            media_list.append(f"#{player}: {media}")

                if len(media_list) > 8:
                    media = " ".join(media_list)
                else:
                    media = "\n".join(media_list)

            if pj[cfg.OBSERVATIONS][obs][cfg.TYPE] == cfg.LIVE:
                media = cfg.LIVE

            if pj[cfg.OBSERVATIONS][obs][cfg.TYPE] == cfg.IMAGES:
                dir_list: list = []
                for dir_path in pj[cfg.OBSERVATIONS][obs].get(cfg.DIRECTORIES_LIST, []):
                    dir_list.append(dir_path)
                media = "; ".join(dir_list)

            # independent variables
            indepvar: list = []
            if cfg.INDEPENDENT_VARIABLES in pj[cfg.OBSERVATIONS][obs]:
                for var_label in indep_var_header:
                    if var_label in pj[cfg.OBSERVATIONS][obs][cfg.INDEPENDENT_VARIABLES]:
                        indepvar.append(pj[cfg.OBSERVATIONS][obs][cfg.INDEPENDENT_VARIABLES][var_label])
                    else:
                        indepvar.append("")

            # check unpaired events
            ok, _ = project_functions.check_state_events_obs(obs, pj[cfg.ETHOGRAM], pj[cfg.OBSERVATIONS][obs], cfg.HHMMSS)
            if not ok:
                not_paired.append(obs)

            # exhaustivity
            if pj[cfg.OBSERVATIONS][obs][cfg.TYPE] in (cfg.MEDIA, cfg.LIVE):
                # check exhaustivity of observation
                exhaustivity = project_functions.check_observation_exhaustivity(pj[cfg.OBSERVATIONS][obs][cfg.EVENTS], state_events_list)
            elif pj[cfg.OBSERVATIONS][obs][cfg.TYPE] == cfg.IMAGES:
                exhaustivity = project_functions.check_observation_exhaustivity_pictures(pj[cfg.OBSERVATIONS][obs])

            data.append([obs, date, descr, subjectsList, observed_interval_str, str(exhaustivity), media] + indepvar)

        obsList = observations_list.observationsList_widget(
            data, header=fields_list + indep_var_header, column_type=column_type, not_paired=not_paired
        )
        self.data = data
        self.not_paired = not_paired
        self.mem_hash_obs = hash(str(self.pj[cfg.OBSERVATIONS]))

    else:
        obsList = observations_list.observationsList_widget(
            self.data, header=fields_list + indep_var_header, column_type=column_type, not_paired=self.not_paired
        )

    if windows_title:
        obsList.setWindowTitle(windows_title)

    obsList.pbOpen.setVisible(False)
    obsList.pbView.setVisible(False)
    obsList.pbEdit.setVisible(False)
    obsList.pbOk.setVisible(False)
    obsList.pbSelectAll.setVisible(False)
    obsList.pbUnSelectAll.setVisible(False)
    obsList.mode = mode

    if mode == cfg.OPEN:
        obsList.view.setSelectionMode(QAbstractItemView.SingleSelection)
        obsList.pbOpen.setVisible(True)

    if mode == cfg.VIEW:
        obsList.view.setSelectionMode(QAbstractItemView.SingleSelection)
        obsList.pbView.setVisible(True)

    if mode == cfg.EDIT:
        obsList.view.setSelectionMode(QAbstractItemView.SingleSelection)
        obsList.pbEdit.setVisible(True)

    if mode == cfg.SINGLE:
        obsList.view.setSelectionMode(QAbstractItemView.SingleSelection)
        obsList.pbOpen.setVisible(True)
        obsList.pbView.setVisible(True)
        obsList.pbEdit.setVisible(True)

    if mode == cfg.MULTIPLE:
        obsList.view.setSelectionMode(QAbstractItemView.MultiSelection)
        obsList.pbOk.setVisible(True)
        obsList.pbSelectAll.setVisible(True)
        obsList.pbUnSelectAll.setVisible(True)

    if mode == cfg.SELECT1:
        obsList.view.setSelectionMode(QAbstractItemView.SingleSelection)
        obsList.pbOk.setVisible(True)

    # restore window geometry
    gui_utilities.restore_geometry(obsList, "observations list", (900, 600))

    obsList.view.sortItems(0, Qt.AscendingOrder)
    for row in range(obsList.view.rowCount()):
        obsList.view.resizeRowToContents(row)

    selected_observations = []

    result = obsList.exec_()

    # saving window geometry in ini file
    gui_utilities.save_geometry(obsList, "observations list")

    if result:
        if obsList.view.selectedIndexes():
            for idx in obsList.view.selectedIndexes():
                if idx.column() == 0:  # first column
                    selected_observations.append(idx.data())

    if result == 0:  # cancel
        resultStr = ""
    if result == 1:  # select
        resultStr = "ok"
    if result == 2:  # open
        resultStr = cfg.OPEN
    if result == 3:  # edit
        resultStr = cfg.EDIT
    if result == 4:  # view
        resultStr = cfg.VIEW

    return resultStr, selected_observations



"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2025 Olivier Friard

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
  MA 02110-1301, USA.
"""

import os
import pathlib
from decimal import Decimal as dec

import tablib
from PySide6.QtWidgets import QFileDialog, QInputDialog, QMessageBox

from . import observation_operations

from . import dialog
from . import project_functions
from . import select_observations
from . import utilities as util
from . import config as cfg
from . import select_subj_behav


def create_behavior_binary_table(pj: dict, selected_observations: list, parameters_obs: dict, time_interval: float) -> dict:
    """
    create behavior binary table

    Args:
        pj (dict): project dictionary
        selected_observations (list): list of selected observations
        parameters_obs (dict): dcit of parameters
        time_interval (float): time interval (in seconds)

    Returns:
        dict: dictionary of tablib dataset

    """

    results_df = {}

    state_behavior_codes = [x for x in util.state_behavior_codes(pj[cfg.ETHOGRAM]) if x in parameters_obs[cfg.SELECTED_BEHAVIORS]]
    point_behavior_codes = [x for x in util.point_behavior_codes(pj[cfg.ETHOGRAM]) if x in parameters_obs[cfg.SELECTED_BEHAVIORS]]
    if not state_behavior_codes and not point_behavior_codes:
        return {"error": True, "msg": "No events selected"}

    for obs_id in selected_observations:
        start_time = parameters_obs[cfg.START_TIME]
        end_time = parameters_obs[cfg.END_TIME]

        # check observation interval
        if parameters_obs["time"] == cfg.TIME_FULL_OBS:
            max_obs_length, _ = observation_operations.observation_length(pj, [obs_id])
            start_time = dec("0.000")
            end_time = dec(max_obs_length)

        if parameters_obs["time"] == cfg.TIME_EVENTS:
            try:
                start_time = dec(pj[cfg.OBSERVATIONS][obs_id][cfg.EVENTS][0][0])
            except Exception:
                start_time = dec("0.000")
            try:
                end_time = dec(pj[cfg.OBSERVATIONS][obs_id][cfg.EVENTS][-1][0])
            except Exception:
                max_obs_length, _ = observation_operations.observation_length(pj, [obs_id])
                end_time = dec(max_obs_length)

        if parameters_obs["time"] == cfg.TIME_OBS_INTERVAL:
            obs_interval = pj[cfg.OBSERVATIONS][obs_id].get(cfg.OBSERVATION_TIME_INTERVAL, [0, 0])
            offset = pj[cfg.OBSERVATIONS][obs_id][cfg.TIME_OFFSET]
            start_time = dec(obs_interval[0]) + offset
            # Use max observation length for end time if no interval is defined (=0)
            max_obs_length, _ = observation_operations.observation_length(pj, [obs_id])
            end_time = dec(obs_interval[1]) + offset if obs_interval[1] not in (0, None) else dec(max_obs_length)

        if obs_id not in results_df:
            results_df[obs_id] = {}

        for subject in parameters_obs[cfg.SELECTED_SUBJECTS]:
            # extract tuple (behavior, modifier)
            behav_modif_list = [
                (idx[2], idx[3])
                for idx in pj[cfg.OBSERVATIONS][obs_id][cfg.EVENTS]
                if idx[1] == (subject if subject != cfg.NO_FOCAL_SUBJECT else "") and idx[2] in parameters_obs[cfg.SELECTED_BEHAVIORS]
            ]

            # extract observed subjects NOT USED at the moment
            """observed_subjects = [
                event[cfg.EVENT_SUBJECT_FIELD_IDX] for event in pj[cfg.OBSERVATIONS][obs_id][cfg.EVENTS]
            ]"""

            # add selected behavior if not found in (behavior, modifier)
            if not parameters_obs[cfg.EXCLUDE_BEHAVIORS]:
                # for behav in state_behavior_codes:
                for behav in parameters_obs[cfg.SELECTED_BEHAVIORS]:
                    if behav not in [x[0] for x in behav_modif_list]:
                        behav_modif_list.append((behav, ""))

            behav_modif_set = set(behav_modif_list)
            observed_behav = [(x[0], x[1]) for x in sorted(behav_modif_set)]
            if parameters_obs[cfg.INCLUDE_MODIFIERS]:
                results_df[obs_id][subject] = tablib.Dataset(
                    headers=["time"] + [f"{x[0]}" + f" ({x[1]})" * (x[1] != "") for x in sorted(behav_modif_set)]
                )
            else:
                results_df[obs_id][subject] = tablib.Dataset(headers=["time"] + [x[0] for x in sorted(behav_modif_set)])

            if subject == cfg.NO_FOCAL_SUBJECT:
                sel_subject_dict = {"": {cfg.SUBJECT_NAME: ""}}
            else:
                sel_subject_dict = dict(
                    [(idx, pj[cfg.SUBJECTS][idx]) for idx in pj[cfg.SUBJECTS] if pj[cfg.SUBJECTS][idx][cfg.SUBJECT_NAME] == subject]
                )

            row_idx = 0
            t = start_time
            while t <= end_time:
                # state events
                current_states = util.get_current_states_modifiers_by_subject_2(
                    state_behavior_codes, pj[cfg.OBSERVATIONS][obs_id][cfg.EVENTS], sel_subject_dict, t
                )

                # point events
                current_point = util.get_current_points_by_subject(
                    point_behavior_codes, pj[cfg.OBSERVATIONS][obs_id][cfg.EVENTS], sel_subject_dict, t, time_interval
                )

                cols = [float(t)]  # time

                for behav in observed_behav:
                    if behav[0] in state_behavior_codes:
                        cols.append(int(behav in current_states[list(current_states.keys())[0]]))

                    if behav[0] in point_behavior_codes:
                        cols.append(current_point[list(current_point.keys())[0]].count(behav))

                results_df[obs_id][subject].append(cols)

                t += time_interval
                row_idx += 1

    return results_df


def behavior_binary_table(self):
    """
    ask user for parameters for behavior binary table
    call create_behavior_binary_table
    """

    QMessageBox.warning(
        None,
        cfg.programName,
        (
            "Depending on the length of yours observations "
            "the execution of this function may take a long time.<br>"
            "The program interface may freeze, be patient. <br>"
        ),
    )

    _, selected_observations = select_observations.select_observations2(
        self, cfg.MULTIPLE, "Select observations for the behavior binary table"
    )

    if not selected_observations:
        return

    # check if coded behaviors are defined in ethogram
    if project_functions.check_coded_behaviors_in_obs_list(self.pj, selected_observations):
        return

    # check if state events are paired
    not_ok, selected_observations = project_functions.check_state_events(self.pj, selected_observations)
    if not_ok or not selected_observations:
        return

    """
    max_obs_length, _ = observation_operations.observation_length(self.pj, selected_observations)
    if max_obs_length == dec(-1):  # media length not available, user choose to not use events
        return

    # exit with message if events do not have timestamp
    if max_obs_length.is_nan():
        QMessageBox.critical(
            None,
            cfg.programName,
            ("This function is not available for observations with events that do not have timestamp"),
            QMessageBox.Ok | QMessageBox.Default,
            QMessageBox.NoButton,
        )
        return
    """

    max_media_duration_all_obs, _ = observation_operations.media_duration(self.pj[cfg.OBSERVATIONS], selected_observations)

    start_coding, end_coding, _ = observation_operations.coding_time(self.pj[cfg.OBSERVATIONS], selected_observations)

    start_interval, end_interval = observation_operations.time_intervals_range(self.pj[cfg.OBSERVATIONS], selected_observations)

    parameters = select_subj_behav.choose_obs_subj_behav_category(
        self,
        selected_observations,
        start_coding=start_coding,
        end_coding=end_coding,
        start_interval=start_interval,
        end_interval=end_interval,
        maxTime=max_media_duration_all_obs,
        show_include_modifiers=True,
        show_exclude_non_coded_behaviors=True,
        by_category=False,
        n_observations=len(selected_observations),
    )
    if not parameters:
        return
    if not parameters[cfg.SELECTED_SUBJECTS] or not parameters[cfg.SELECTED_BEHAVIORS]:
        QMessageBox.warning(None, cfg.programName, "Select subject(s) and behavior(s) to analyze")
        return

    # ask for time interval
    i, ok = QInputDialog.getDouble(None, "Behavior binary table", "Time interval (in seconds):", 1.0, 0.001, 86400, 3)
    if not ok:
        return
    time_interval = util.float2decimal(i)

    results_df = create_behavior_binary_table(self.pj, selected_observations, parameters, time_interval)

    if "error" in results_df:
        QMessageBox.warning(None, cfg.programName, results_df["msg"])
        return

    # save results
    file_formats = [cfg.TSV, cfg.CSV, cfg.ODS, cfg.XLSX, cfg.XLS, cfg.HTML]

    if len(selected_observations) == 1:
        file_name, filter_ = QFileDialog().getSaveFileName(None, "Save results", "", ";;".join(file_formats))
        if not file_name:
            return

        output_format = cfg.FILE_NAME_SUFFIX[filter_]

        if pathlib.Path(file_name).suffix != "." + output_format:
            file_name = str(pathlib.Path(file_name)) + "." + output_format
            # check if file with new extension already exists
            if pathlib.Path(file_name).is_file():
                if (
                    dialog.MessageDialog(cfg.programName, f"The file {file_name} already exists.", [cfg.CANCEL, cfg.OVERWRITE])
                    == cfg.CANCEL
                ):
                    return
    else:
        item, ok = QInputDialog.getItem(None, "Save results", "Available formats", file_formats, 0, False)
        if not ok:
            return

        output_format = cfg.FILE_NAME_SUFFIX[item]

        export_dir = QFileDialog().getExistingDirectory(
            None, "Choose a directory to save results", os.path.expanduser("~"), options=QFileDialog.ShowDirsOnly
        )
        if not export_dir:
            return

    mem_command = ""
    for obs_id in results_df:
        for subject in results_df[obs_id]:
            if len(selected_observations) > 1:
                file_name_with_subject = str(pathlib.Path(export_dir) / util.safeFileName(obs_id + "_" + subject)) + "." + output_format
            else:
                file_name_with_subject = str(os.path.splitext(file_name)[0] + util.safeFileName("_" + subject)) + "." + output_format

            # check if file with new extension already exists
            if mem_command != cfg.OVERWRITE_ALL and pathlib.Path(file_name_with_subject).is_file():
                if mem_command == "Skip all":
                    continue
                mem_command = dialog.MessageDialog(
                    cfg.programName,
                    f"The file {file_name_with_subject} already exists.",
                    [cfg.OVERWRITE, cfg.OVERWRITE_ALL, "Skip", "Skip all", cfg.CANCEL],
                )
                if mem_command == cfg.CANCEL:
                    return
                if mem_command in ["Skip", "Skip all"]:
                    continue

            if output_format in [cfg.CSV_EXT, cfg.TSV_EXT, cfg.HTML]:
                with open(file_name_with_subject, "wb") as f:
                    f.write(str.encode(results_df[obs_id][subject].export(output_format)))

            if output_format in [cfg.ODS_EXT, cfg.XLSX_EXT, cfg.XLS_EXT]:
                with open(file_name_with_subject, "wb") as f:
                    f.write(results_df[obs_id][subject].export(output_format))



"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2025 Olivier Friard

This file is part of BORIS.

  BORIS is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  any later version.

  BORIS is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not see <http://www.gnu.org/licenses/>.

"""

import binascii
import io
import json
from pathlib import Path
import gui_utilities

from PySide6.QtCore import QBuffer, QByteArray, QIODevice, QLineF, QPoint, Qt, Signal
from PySide6.QtGui import QBrush, QColor, QIcon, QMouseEvent, QPen, QPixmap, QPolygonF, QAction
from PySide6.QtWidgets import (
    QApplication,
    QColorDialog,
    QFileDialog,
    QFrame,
    QGraphicsEllipseItem,
    QGraphicsLineItem,
    QGraphicsPixmapItem,
    QGraphicsPolygonItem,
    QGraphicsScene,
    QGraphicsView,
    QHBoxLayout,
    QInputDialog,
    QLabel,
    QLineEdit,
    QListWidget,
    QMainWindow,
    QMessageBox,
    QPushButton,
    QSizePolicy,
    QSlider,
    QSpacerItem,
    QSplitter,
    QVBoxLayout,
    QWidget,
)

from . import config as cfg
from . import dialog
from . import utilities as util

designColor = QColor(255, 0, 0, 128)  # red opacity: 50%
penWidth = 0
penStyle = Qt.NoPen
selectedBrush = QBrush()
selectedBrush.setStyle(Qt.SolidPattern)
selectedBrush.setColor(QColor(255, 255, 0, 255))


class BehaviorsMapCreatorWindow(QMainWindow):
    signal_add_to_project = Signal(dict)

    class View(QGraphicsView):
        """
        class for handling mousepress event in QGraphicsView
        """

        mousePress = Signal(QMouseEvent)

        def mousePressEvent(self, event):
            self.mousePress.emit(event)

        _start = 0
        elList, points = [], []

        def __init__(self, parent):
            QGraphicsView.__init__(self, parent)
            self.setBackgroundBrush(QColor(128, 128, 128))
            self.setScene(QGraphicsScene(self))
            self.scene().update()

    bitmapFileName, mapName, fileName = "", "", ""
    flagNewArea, flag_map_changed = False, False
    polygonsList2 = []
    areaColor = QColor("lime")

    def __init__(self, arg):
        self.codes_list = arg

        super(BehaviorsMapCreatorWindow, self).__init__()

        self.pixmap = QPixmap()
        self.closedPolygon = None
        self.selectedPolygon = None

        self.setWindowTitle("BORIS - Behaviors coding map creator")

        self.newMapAction = QAction(QIcon(), "&New behaviors coding map", self)
        self.newMapAction.setShortcut("Ctrl+N")
        self.newMapAction.setStatusTip("Create a new behaviors coding map")
        self.newMapAction.triggered.connect(self.newMap)

        self.openMapAction = QAction(QIcon(), "&Open a behaviors coding map", self)
        self.openMapAction.setShortcut("Ctrl+O")
        self.openMapAction.setStatusTip("Open a behaviors coding map")
        self.openMapAction.triggered.connect(self.openMap)

        self.saveMapAction = QAction(QIcon(), "&Save the behavior coding map", self)
        self.saveMapAction.setShortcut("Ctrl+S")
        self.saveMapAction.setStatusTip("Save the behavior coding map")
        self.saveMapAction.setEnabled(False)
        self.saveMapAction.triggered.connect(self.saveMap_clicked)

        self.saveAsMapAction = QAction(QIcon(), "Save the behavior coding map as ...", self)
        self.saveAsMapAction.setStatusTip("Save the behavior coding map as ...")
        self.saveAsMapAction.setEnabled(False)
        self.saveAsMapAction.triggered.connect(self.saveAsMap_clicked)

        self.mapNameAction = QAction(QIcon(), "&Edit name of behaviors coding map", self)
        self.mapNameAction.setShortcut("Ctrl+M")
        self.mapNameAction.setStatusTip("Edit name of behaviors coding map")
        self.mapNameAction.setEnabled(False)
        self.mapNameAction.triggered.connect(self.mapName_clicked)

        self.resizeAction = QAction(QIcon(), "Resize the image", self)
        self.resizeAction.setStatusTip("Resize the image")
        self.resizeAction.setEnabled(False)
        self.resizeAction.triggered.connect(self.resize_clicked)

        self.addToProject = QAction(QIcon(), "Add coding map to project", self)
        self.addToProject.setStatusTip("Add coding map to project")
        self.addToProject.setEnabled(False)
        self.addToProject.triggered.connect(self.add_to_project)

        self.exitAction = QAction(QIcon(), "&Close", self)
        self.exitAction.setStatusTip(cfg.CLOSE)
        self.exitAction.triggered.connect(self.close)

        menubar = self.menuBar()
        fileMenu = menubar.addMenu("&File")
        fileMenu.addAction(self.newMapAction)
        fileMenu.addAction(self.openMapAction)
        fileMenu.addAction(self.saveMapAction)
        fileMenu.addAction(self.saveAsMapAction)
        fileMenu.addSeparator()
        fileMenu.addAction(self.mapNameAction)
        fileMenu.addSeparator()
        fileMenu.addAction(self.resizeAction)
        fileMenu.addSeparator()
        fileMenu.addAction(self.addToProject)
        fileMenu.addSeparator()
        fileMenu.addAction(self.exitAction)

        splitter1 = QSplitter(Qt.Vertical)

        self.view = self.View(self)
        self.view.mousePress.connect(self.viewMousePressEvent)
        splitter1.addWidget(self.view)

        vlayout_list = QVBoxLayout()
        vlayout_list.addWidget(QLabel("Defined area"))

        self.area_list = QListWidget(self)
        self.area_list.itemClicked.connect(self.area_list_item_click)
        vlayout_list.addWidget(self.area_list)
        w = QWidget()
        w.setLayout(vlayout_list)
        splitter1.addWidget(w)
        splitter1.setSizes([300, 100])
        splitter1.setStretchFactor(2, 8)

        hlayout_cmd = QHBoxLayout()

        self.btNewArea = QPushButton("New behavior area", self)
        self.btNewArea.clicked.connect(self.newArea)
        self.btNewArea.setVisible(False)
        hlayout_cmd.addWidget(self.btNewArea)

        self.btSaveArea = QPushButton("Save the behavior area", self)
        self.btSaveArea.clicked.connect(self.saveArea)
        self.btSaveArea.setVisible(False)
        hlayout_cmd.addWidget(self.btSaveArea)

        self.btCancelAreaCreation = QPushButton("Cancel", self)
        self.btCancelAreaCreation.clicked.connect(self.cancelAreaCreation)
        self.btCancelAreaCreation.setVisible(False)
        hlayout_cmd.addWidget(self.btCancelAreaCreation)

        hlayout_cmd.addItem(QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum))

        self.btDeleteArea = QPushButton("Delete selected behavior area", self)
        self.btDeleteArea.clicked.connect(self.deleteArea)
        self.btDeleteArea.setVisible(False)
        hlayout_cmd.addWidget(self.btDeleteArea)

        hlayout_area = QHBoxLayout()

        self.lb = QLabel("Behavior")
        self.lb.setVisible(False)
        hlayout_area.addWidget(self.lb)

        self.leAreaCode = QLineEdit(self)
        self.leAreaCode.setReadOnly(True)
        self.leAreaCode.setVisible(False)
        self.leAreaCode.setEnabled(False)
        hlayout_area.addWidget(self.leAreaCode)

        self.btEditAreaCode = QPushButton("Select behavior")
        self.btEditAreaCode.clicked.connect(self.edit_area_code)
        self.btEditAreaCode.setVisible(False)
        hlayout_area.addWidget(self.btEditAreaCode)

        self.btColor = QPushButton(clicked=self.chooseColor)
        self.btColor.setVisible(False)
        self.btColor.setStyleSheet(f"QWidget {{background-color:{self.areaColor.name()}}}")
        hlayout_area.addWidget(self.btColor)

        self.slAlpha = QSlider(Qt.Horizontal)
        self.slAlpha.setRange(20, 100)
        self.slAlpha.setValue(50)
        self.slAlpha.valueChanged.connect(self.slAlpha_changed)
        self.slAlpha.setVisible(False)
        hlayout_area.addWidget(self.slAlpha)

        self.slAlpha_changed(50)

        vlayout_frame = QVBoxLayout()
        vlayout_frame.addLayout(hlayout_cmd)
        vlayout_frame.addLayout(hlayout_area)
        vlayout_frame.addItem(QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding))

        frame = QFrame()
        frame.setFrameStyle(QFrame.Panel | QFrame.Plain)
        frame.setMinimumHeight(120)
        frame.setMaximumHeight(120)

        frame.setLayout(vlayout_frame)

        vlayout = QVBoxLayout()

        vlayout.addWidget(splitter1)
        vlayout.addWidget(frame)

        main_widget = QWidget(self)
        main_widget.setLayout(vlayout)
        self.setCentralWidget(main_widget)

        self.statusBar().showMessage("")

    def add_to_project(self, item):
        """
        add coding map to project
        """

        map_dict = self.make_coding_map_dict()
        if map_dict["areas"] == {}:
            QMessageBox.critical(
                self,
                cfg.programName,
                (
                    "The behaviors coding map does not contain any behavior area.<br>"
                    "Add some behavior areas before adding the coding map to the project"
                ),
            )

            return

        self.signal_add_to_project.emit(map_dict)

    def area_list_item_click(self, item):
        """
        select the polygon corresponding to the clicked area
        """

        if self.selectedPolygon:
            self.selectedPolygon.setPen(QPen(designColor, penWidth, penStyle, Qt.RoundCap, Qt.RoundJoin))
            self.selectedPolygon = None
            self.selectedPolygonMemBrush = None

        idx = int(item.text().split("#")[1])
        ac, pg = self.polygonsList2[idx]

        self.selectedPolygon = pg

        self.selectedPolygonMemBrush = self.selectedPolygon.brush()

        self.selectedPolygon.setPen(QPen(QColor(255, 0, 0, 255), 2, Qt.SolidLine, Qt.RoundCap, Qt.RoundJoin))

        self.leAreaCode.setText(ac)

        for widget in (
            self.lb,
            self.leAreaCode,
            self.btEditAreaCode,
            self.btDeleteArea,
            self.btColor,
            self.slAlpha,
        ):
            widget.setVisible(True)

        self.areaColor = self.selectedPolygon.brush().color()
        self.btColor.setStyleSheet(f"QWidget {{background-color:{self.selectedPolygon.brush().color().name()}}}")
        self.slAlpha.setValue(int(self.areaColor.alpha() / 255 * 100))

    def edit_area_code(self):
        """
        select a behavior
        """

        if self.leAreaCode.text() in self.codes_list:
            code_index = self.codes_list.index(self.leAreaCode.text())
        else:
            code_index = 0

        item, ok = QInputDialog.getItem(self, "Select a behavior", "Available behaviors", self.codes_list, code_index, False)
        self.leAreaCode.setText(item)

        if self.selectedPolygon:
            for idx, area in enumerate(self.polygonsList2):
                ac, pg = area
                if pg == self.selectedPolygon:
                    self.polygonsList2[idx] = [self.leAreaCode.text(), pg]
                    break

            self.update_area_list()

    def slAlpha_changed(self, val):
        """
        opacity slider value changed
        """

        self.btColor.setText(f"Opacity: {val} %")
        self.areaColor.setAlpha(int(val / 100 * 255))

        if self.selectedPolygon:
            self.selectedPolygon.setBrush(self.areaColor)
            for idx, area in enumerate(self.polygonsList2):
                ac, pg = area
                if pg == self.selectedPolygon:
                    pg.setBrush(self.areaColor)
                    self.polygonsList2[idx] = [ac, pg]
                    break

        if self.closedPolygon:
            self.closedPolygon.setBrush(self.areaColor)

    def chooseColor(self):
        """
        area color button clicked
        """
        cd = QColorDialog()
        cd.setWindowFlags(Qt.WindowStaysOnTopHint)
        cd.setOptions(QColorDialog.DontUseNativeDialog)

        if cd.exec():
            self.areaColor = cd.currentColor()
            self.btColor.setStyleSheet(f"QWidget {{background-color:{self.areaColor.name()}}}")
            self.areaColor.setAlpha(int(self.slAlpha.value() / 100 * 255))

        if self.selectedPolygon:
            self.selectedPolygon.setBrush(self.areaColor)

            for idx, area in enumerate(self.polygonsList2):
                ac, pg = area
                if pg == self.selectedPolygon:
                    pg.setBrush(self.areaColor)
                    self.polygonsList2[idx] = [ac, pg]
                    break

        if self.closedPolygon:
            self.closedPolygon.setBrush(self.areaColor)

    def closeEvent(self, event):
        if self.flag_map_changed:
            response = dialog.MessageDialog(
                "BORIS - Behaviors map creator",
                "What to do about the current unsaved behaviors coding map?",
                [cfg.SAVE, cfg.DISCARD, cfg.CANCEL],
            )

            if response == cfg.SAVE:
                if not self.saveMap_clicked():
                    event.ignore()

            if response == cfg.CANCEL:
                event.ignore()
                return

        self.flag_map_changed = False
        event.accept()

    def viewMousePressEvent(self, event):
        """
        check if area selected with mouse
        """

        def add_polygon():
            # create polygon
            newPolygon = QPolygonF()
            for p in self.view.points:
                newPolygon.append(QPoint(p[0], p[1]))

            # draw polygon a red polygon
            self.closedPolygon = QGraphicsPolygonItem(newPolygon)
            self.closedPolygon.setPen(QPen(designColor, penWidth, penStyle, Qt.RoundCap, Qt.RoundJoin))
            self.closedPolygon.setBrush(self.areaColor)
            self.view.scene().addItem(self.closedPolygon)

        if not self.bitmapFileName:
            return

        test = self.view.mapToScene(event.pos()).toPoint()  # coordinates of clicked point

        if test.x() < 0 or test.y() < 0 or test.x() > self.pixmap.size().width() or test.y() > self.pixmap.size().height():
            return

        if not self.flagNewArea:  # test clicked point for areas
            # reset selected polygon to default pen
            if self.selectedPolygon:
                self.selectedPolygon.setPen(QPen(designColor, penWidth, penStyle, Qt.RoundCap, Qt.RoundJoin))
                self.selectedPolygon = None
                self.selectedPolygonMemBrush = None

            print(self.polygonsList2)
            idx = 0
            for areaCode, pg in self.polygonsList2:
                if pg.contains(test):
                    self.selectedPolygon = pg
                    self.selectedPolygonMemBrush = self.selectedPolygon.brush()
                    self.selectedPolygon.setPen(QPen(QColor(255, 0, 0, 255), 2, Qt.SolidLine, Qt.RoundCap, Qt.RoundJoin))
                    self.leAreaCode.setText(areaCode)

                    for widget in (self.lb, self.leAreaCode, self.btEditAreaCode, self.btColor, self.slAlpha, self.btDeleteArea):
                        widget.setVisible(True)

                    self.areaColor = self.selectedPolygon.brush().color()
                    self.btColor.setStyleSheet(f"QWidget {{background-color:{self.selectedPolygon.brush().color().name()}}}")

                    self.slAlpha.setValue(int(self.areaColor.alpha() / 255 * 100))

                    # select area in list widget
                    item = self.area_list.item(idx)
                    self.area_list.setCurrentItem(item)

                    break
                idx += 1

            if not self.selectedPolygon:
                self.leAreaCode.setVisible(False)
                self.lb.setVisible(False)
                self.btDeleteArea.setVisible(False)
                self.btEditAreaCode.setVisible(False)
                self.btColor.setVisible(False)
                self.slAlpha.setVisible(False)
            return

        # delete last line item
        if (event.buttons() & Qt.RightButton) and not self.closedPolygon:
            if self.view.points:
                self.view.points = self.view.points[0:-1]

            if self.view.points:
                self.view._start = QPoint(self.view.points[-1][0], self.view.points[-1][1])
            else:
                self.view._start = None

            # remove graphical elements
            if self.view.elList:
                self.view.scene().removeItem(self.view.elList[-1])
                self.view.elList = self.view.elList[0:-1]

        # middle button automatically close the polygon
        if (event.buttons() & Qt.MiddleButton) and not self.closedPolygon:
            line = QGraphicsLineItem(QLineF(self.view._start, QPoint(self.view.points[0][0], self.view.points[0][1])))
            line.setPen(QPen(designColor, penWidth, Qt.SolidLine, Qt.RoundCap, Qt.RoundJoin))

            self.view.scene().addItem(line)
            self.view.elList.append(line)

            self.statusBar().showMessage("Area completed")

            # create polygon
            add_polygon()
            return

        # add line item
        if event.buttons() == Qt.LeftButton and not self.closedPolygon:
            if self.view._start:
                end = test

                # test is polygon is crossed
                if len(self.view.points) >= 3:
                    for idx, _ in enumerate(self.view.points[:-2]):
                        if util.intersection(
                            self.view.points[idx],
                            self.view.points[idx + 1],
                            self.view.points[-1],
                            (int(end.x()), int(end.y())),
                        ):
                            QMessageBox.critical(self, "", "The polygon edges can not be intersected")
                            return

                # test if polygon closed (dist min 10 px)
                if abs(end.x() - self.view.points[0][0]) < 10 and abs(end.y() - self.view.points[0][1]) < 10:
                    line = QGraphicsLineItem(QLineF(self.view._start, QPoint(self.view.points[0][0], self.view.points[0][1])))
                    line.setPen(QPen(designColor, penWidth, Qt.SolidLine, Qt.RoundCap, Qt.RoundJoin))

                    self.view.scene().addItem(line)
                    self.view.elList.append(line)

                    self.statusBar().showMessage("Area completed")

                    # create polygon
                    add_polygon()
                    return

                self.view.points.append((int(end.x()), int(end.y())))

                line = QGraphicsLineItem(QLineF(self.view._start, end))
                line.setPen(QPen(designColor, 2, Qt.SolidLine, Qt.RoundCap, Qt.RoundJoin))
                self.view.scene().addItem(line)
                self.view.elList.append(line)

                self.view._start = test

            else:  # first point
                self.view._start = test

                ellipse = QGraphicsEllipseItem(self.view._start.x(), self.view._start.y(), 3, 3)
                ellipse.setPen(QPen(designColor, 1, Qt.SolidLine, Qt.RoundCap, Qt.RoundJoin))

                brush = QBrush()
                brush.setStyle(Qt.SolidPattern)
                brush.setColor(designColor)
                ellipse.setBrush(brush)

                self.view.scene().addItem(ellipse)
                self.view.elList.append(ellipse)

                self.view.points.append((self.view._start.x(), self.view._start.y()))

        # automatically close the polygon
        if event.buttons() == Qt.MiddleButton and not self.closedPolygon:
            # add first point as last point of polygon

            # test is polygon is crossed

            if len(self.view.points) >= 3:
                for idx, _ in enumerate(self.view.points[1:-2]):
                    if util.intersection(
                        self.view.points[idx],
                        self.view.points[idx + 1],
                        self.view.points[-1],
                        self.view.points[0],
                    ):
                        QMessageBox.critical(self, "", "The polygon edges can not be intersected")
                        return

            line = QGraphicsLineItem(
                QLineF(self.view.points[-1][0], self.view.points[-1][1], self.view.points[0][0], self.view.points[0][1])
            )
            line.setPen(QPen(designColor, 2, Qt.SolidLine, Qt.RoundCap, Qt.RoundJoin))
            self.view.scene().addItem(line)
            self.view.elList.append(line)

            self.view.points.append(self.view.points[0])
            add_polygon()
            self.statusBar().showMessage("Area completed")

    def mapName_clicked(self):
        """
        Edit map name
        """

        while True:
            map_name, ok = QInputDialog.getText(
                self, "Behaviors coding map name", "Enter a name for the coding map", QLineEdit.Normal, self.mapName
            )
            if map_name.upper() in self.bcm_list:
                QMessageBox.critical(
                    self,
                    "",
                    (
                        "The name for the new coding map already exists.<br>"
                        f"{', '.join(self.bcm_list)} are already defined.<br>"
                        "To reuse the same name the existing coding map must be deleted (File > Edit project)"
                    ),
                )
            if ok and map_name and map_name.upper() not in self.bcm_list:
                self.mapName = map_name
                self.setWindowTitle(f"{cfg.programName} - Behaviors coding map creator - {self.mapName}")
                break
            if not ok:
                return

    def newMap(self):
        """
        create a new map
        """

        if self.flag_map_changed:
            response = dialog.MessageDialog(
                cfg.programName + " - Behaviors coding map creator",
                "What to do about the current unsaved coding map?",
                [cfg.SAVE, cfg.DISCARD, cfg.CANCEL],
            )

            if response == cfg.SAVE:
                if not self.saveMap_clicked():
                    return

            if response == cfg.CANCEL:
                return

        self.cancelMap()

        while True:
            map_name, ok = QInputDialog.getText(self, "Behaviors coding map name", "Enter a name for the new coding map")
            if not ok:
                return
            if map_name.upper() in self.bcm_list:
                QMessageBox.critical(
                    self,
                    "",
                    (
                        "The name for the new coding map already exists.<br>"
                        f"{', '.join(self.bcm_list)} are already defined.<br>"
                        "To reuse the same name the existing coding map must be deleted (File > Edit project)"
                    ),
                )
            if ok and map_name and map_name.upper() not in self.bcm_list:
                self.mapName = map_name
                break

        self.setWindowTitle(f"{cfg.programName} - Behaviors coding map creator tool - {self.mapName}")

        self.loadBitmap()

    def openMap(self):
        """
        open a coding map from file

        load bitmap from data
        show it in view scene
        """
        if self.flag_map_changed:
            response = dialog.MessageDialog(
                cfg.programName + " - Behaviors coding map creator",
                "What to do about the current unsaved coding map?",
                ["Save", "Discard", "Cancel"],
            )

            if (response == "Save" and not self.saveMap_clicked()) or (response == "Cancel"):
                return

        fileName, _ = QFileDialog(self).getOpenFileName(
            self, "Open a behaviors coding map", "", "Behaviors coding map (*.behav_coding_map);;All files (*)"
        )

        if not fileName:
            return
        try:
            self.codingMap = json.loads(open(fileName, "r").read())
        except Exception:
            QMessageBox.critical(self, cfg.programName, f"The file {fileName} is not a behaviors coding map.")
            return

        if "coding_map_type" not in self.codingMap or self.codingMap["coding_map_type"] != "BORIS behaviors coding map":
            QMessageBox.critical(self, cfg.programName, f"The file {fileName} is not a BORIS behaviors coding map.")
            return

        self.cancelMap()

        self.mapName = self.codingMap["name"]

        self.setWindowTitle(f"{cfg.programName} - Behaviors coding map creator - {self.mapName}")

        self.bitmapFileName = True

        self.fileName = fileName

        bitmapContent = binascii.a2b_base64(self.codingMap["bitmap"])

        self.pixmap.loadFromData(bitmapContent)

        self.view.setSceneRect(0, 0, self.pixmap.size().width(), self.pixmap.size().height())
        self.view.setMinimumHeight(self.pixmap.size().height())
        # self.view.setMaximumHeight(self.pixmap.size().height())
        pixItem = QGraphicsPixmapItem(self.pixmap)
        pixItem.setPos(0, 0)
        self.view.scene().addItem(pixItem)

        for key in self.codingMap["areas"]:
            areaCode = self.codingMap["areas"][key]["code"]
            points = self.codingMap["areas"][key]["geometry"]

            newPolygon = QPolygonF()
            for p in points:
                newPolygon.append(QPoint(p[0], p[1]))

            # draw polygon
            polygon = QGraphicsPolygonItem()
            polygon.setPolygon(newPolygon)
            clr = QColor()
            clr.setRgba(self.codingMap["areas"][key]["color"])
            polygon.setPen(QPen(clr, penWidth, penStyle, Qt.RoundCap, Qt.RoundJoin))
            polygon.setBrush(QBrush(clr, Qt.SolidPattern))

            self.view.scene().addItem(polygon)

            self.polygonsList2.append([areaCode, polygon])

        self.btNewArea.setVisible(True)

        for action in (
            self.saveMapAction,
            self.saveAsMapAction,
            self.addToProject,
            self.mapNameAction,
            self.resizeAction,
        ):
            action.setEnabled(True)

        self.update_area_list()

    def make_coding_map_dict(self) -> dict:
        """
        create the coding map dictionary
        the image is encoded in bse64 format
        """
        map_dict = {"coding_map_type": "BORIS behaviors coding map", "name": self.mapName, "areas": {}}

        for ac, pg in self.polygonsList2:
            if not map_dict["areas"]:
                idx = 0
            else:
                idx = max(map_dict["areas"].keys()) + 1

            points = []
            for p in range(pg.polygon().count()):
                points.append([int(pg.polygon().value(p).x()), int(pg.polygon().value(p).y())])

            map_dict["areas"][idx] = {"code": ac, "geometry": points, "color": pg.brush().color().rgba()}

        map_dict["areas"] = json.loads(json.dumps(map_dict["areas"]))

        # Save QPixmap to QByteArray via QBuffer.
        byte_array = QByteArray()
        buffer = QBuffer(byte_array)
        buffer.open(QIODevice.WriteOnly)
        self.pixmap.save(buffer, "PNG")
        string_io = io.BytesIO(byte_array)
        string_io.seek(0)

        # add codified bitmap
        map_dict["bitmap"] = binascii.b2a_base64(string_io.read()).decode("utf-8")

        return map_dict

    def saveMap(self):
        """
        save current coding map in JSON format
        """

        if self.fileName:
            mapDict = self.make_coding_map_dict()

            with open(self.fileName, "w") as outfile:
                outfile.write(json.dumps(mapDict))

            self.flag_map_changed = False

            return True
        else:
            return False

    def saveAsMap_clicked(self):
        filters = "Behaviors coding map (*.behav_coding_map);;All files (*)"

        self.fileName, _ = QFileDialog.getSaveFileName(self, "Save behaviors coding map as", "", filters)

        if not self.fileName:
            return
        if Path(self.fileName).suffix != ".behav_coding_map":
            self.fileName += ".behav_coding_map"
        self.saveMap()

    def saveMap_clicked(self):
        if not self.fileName:
            self.fileName, _ = QFileDialog().getSaveFileName(
                self,
                "Save modifiers map",
                self.mapName + ".behav_coding_map",
                "Behaviors coding map (*.behav_coding_map);;All files (*)",
            )

            if self.fileName and Path(self.fileName).suffix != ".behav_coding_map":
                self.fileName += ".behav_coding_map"

        if self.fileName:
            return self.saveMap()

        return False

    def newArea(self):
        if not self.bitmapFileName:
            QMessageBox.critical(self, cfg.programName, "An image must be loaded before to define areas")
            return

        if self.selectedPolygon:
            self.selectedPolygon.setPen(QPen(designColor, penWidth, penStyle, Qt.RoundCap, Qt.RoundJoin))
            self.selectedPolygon = None

        self.flagNewArea = True
        self.btSaveArea.setVisible(True)
        self.btCancelAreaCreation.setVisible(True)
        self.btNewArea.setVisible(False)
        self.lb.setVisible(True)
        self.leAreaCode.clear()
        self.leAreaCode.setVisible(True)
        # self.leAreaCode.setEnabled(False)
        self.btEditAreaCode.setVisible(True)
        self.btColor.setVisible(True)
        self.slAlpha.setVisible(True)
        self.btDeleteArea.setVisible(False)

        self.statusBar().showMessage(
            ("Click on bitmap to set the vertices of the area with the mouse (right click will cancel the last point)")
        )

    def saveArea(self):
        if not self.closedPolygon:
            QMessageBox.critical(
                self,
                cfg.programName,
                ("You must close your area before saving it.\nThe last vertex must correspond to the first one."),
            )

        if len(self.view.points) < 3:
            QMessageBox.critical(self, cfg.programName, "You must define a closed area")
            return

        # check if no area code
        if not self.leAreaCode.text():
            QMessageBox.critical(self, cfg.programName, "You must define a code for the new behavior area")
            return

        # remove all lines
        for x in self.view.elList:
            self.view.scene().removeItem(x)

        # draw polygon
        self.closedPolygon.setBrush(QBrush(self.areaColor, Qt.SolidPattern))
        self.polygonsList2.append([self.leAreaCode.text(), self.closedPolygon])

        self.flagNewArea = None
        self.closedPolygon = None
        self.view._start = 0
        self.view.points = []
        self.view.elList = []

        for widget in (
            self.btSaveArea,
            self.btCancelAreaCreation,
            self.lb,
            self.leAreaCode,
            self.btEditAreaCode,
            self.btColor,
            self.slAlpha,
            self.btDeleteArea,
            self.btNewArea,
        ):
            widget.setVisible(False)

        self.btNewArea.setVisible(True)

        self.leAreaCode.setText("")

        self.update_area_list()

        self.flag_map_changed = True
        self.statusBar().showMessage("New area saved", 5000)

    def cancelAreaCreation(self):
        if self.closedPolygon:
            self.view.scene().removeItem(self.closedPolygon)
            self.closedPolygon = None

        # remove all lines
        for x in self.view.elList:
            self.view.scene().removeItem(x)

        self.view.elList = []

        self.view._start = 0
        self.view.points = []
        self.flagNewArea = False
        self.btCancelAreaCreation.setVisible(False)
        self.btDeleteArea.setVisible(False)
        self.btSaveArea.setVisible(False)
        self.lb.setVisible(False)

        self.btColor.setVisible(False)
        self.slAlpha.setVisible(False)
        self.btNewArea.setVisible(True)

        self.leAreaCode.setVisible(False)
        self.leAreaCode.setText("")

        self.btEditAreaCode.setVisible(False)

    def update_area_list(self):
        self.area_list.clear()
        for idx, area in enumerate(self.polygonsList2):
            ac, pg = area
            self.area_list.addItem(f"{ac} #{idx}")

    def deleteArea(self):
        """
        remove selected area from map
        """

        if self.selectedPolygon:
            self.view.scene().removeItem(self.selectedPolygon)

            to_delete = -1
            for idx, area in enumerate(self.polygonsList2):
                ac, pg = area
                if pg == self.selectedPolygon:
                    to_delete = idx

            if to_delete != -1:
                del self.polygonsList2[to_delete]

            self.flag_map_changed = True

        self.view.elList = []

        self.view._start = 0
        self.view.points = []
        self.flagNewArea = False

        for widget in (self.btSaveArea, self.lb, self.btColor, self.slAlpha, self.leAreaCode, self.btDeleteArea):
            widget.setVisible(False)

        self.btNewArea.setVisible(True)

        self.leAreaCode.setText("")
        self.btEditAreaCode.setVisible(False)

        self.statusBar().showMessage("")

        self.update_area_list()

    def cancelMap(self):
        """
        remove current map
        """
        self.flagNewArea = False
        self.polygonsList2 = []
        self.closedPolygon = None
        self.selectedPolygon = None
        self.area_list.clear()
        self.view.scene().clear()
        self.btDeleteArea.setVisible(False)
        self.btNewArea.setVisible(False)
        self.saveMapAction.setEnabled(False)
        self.saveAsMapAction.setEnabled(False)
        self.addToProject.setEnabled(False)
        self.mapNameAction.setEnabled(False)
        self.resizeAction.setEnabled(False)
        self.statusBar().showMessage("")

        self.btNewArea.setVisible(True)
        self.btNewArea.setEnabled(True)

        self.flag_map_changed = False

    def loadBitmap(self):
        """
        load bitmap as background for coding map
        no more resize bitmap to CODING_MAP_RESIZE_W x CODING_MAP_RESIZE_H defined in config.py
        """

        fileName, _ = QFileDialog.getOpenFileName(self, "Load bitmap", "", "bitmap files (*.png *.jpg);;All files (*)")

        if not fileName:
            return
        self.bitmapFileName = fileName

        self.pixmap.load(self.bitmapFileName)

        # scale image
        """
        if (
            self.pixmap.size().width() > cfg.CODING_MAP_RESIZE_W
            or self.pixmap.size().height() > cfg.CODING_MAP_RESIZE_H
        ):
            self.pixmap = self.pixmap.scaled(cfg.CODING_MAP_RESIZE_W, cfg.CODING_MAP_RESIZE_H, Qt.KeepAspectRatio)
            QMessageBox.information(
                self,
                cfg.programName,
                (
                    f"The bitmap was resized to {self.pixmap.size().width()}x{self.pixmap.size().height()} pixels\n"
                    "The original file was not modified"
                ),
            )
        """

        self.view.setSceneRect(0, 0, self.pixmap.size().width(), self.pixmap.size().height())
        pixitem = QGraphicsPixmapItem(self.pixmap)
        pixitem.setPos(0, 0)
        self.view.scene().addItem(pixitem)

        self.btNewArea.setVisible(True)

        self.saveMapAction.setEnabled(True)
        self.saveAsMapAction.setEnabled(True)
        self.addToProject.setEnabled(True)
        self.mapNameAction.setEnabled(True)
        self.resizeAction.setEnabled(True)

        self.statusBar().showMessage("""Click "New behavior area" to create a new behavior area""")

        self.flag_map_changed = True

    def resize_clicked(self):
        """
        resize the bitmap
        """

        if self.polygonsList2:
            if (
                dialog.MessageDialog(
                    "BORIS - Behaviors map creator",
                    "The map contains modifiers. Erase all modifiers",
                    (cfg.NO, cfg.YES),
                )
                == cfg.NO
            ):
                return

        integer, ok = QInputDialog.getInt(
            self, "Resize image", "New horizontal size (in pixels)", value=self.pixmap.size().width(), minValue=100, maxValue=3048, step=10
        )
        if not ok:
            return

        self.cancelMap()

        h = int(self.pixmap.size().width() / (self.pixmap.size().width() / integer))

        self.pixmap = self.pixmap.scaled(integer, h, Qt.KeepAspectRatio)

        self.view.scene().clear()
        self.view.setSceneRect(0, 0, self.pixmap.size().width(), self.pixmap.size().height())
        pixitem = QGraphicsPixmapItem(self.pixmap)
        pixitem.setPos(0, 0)
        self.view.scene().addItem(pixitem)

        QMessageBox.information(
            self,
            cfg.programName,
            (
                f"The bitmap was resized to {self.pixmap.size().width()}x{self.pixmap.size().height()} pixels\n"
                "The original file was not modified"
            ),
        )


def behaviors_coding_map_creator(self):
    """
    show behaviors coding map creator window
    """

    if not self.project:
        QMessageBox.warning(
            self, cfg.programName, "Create or open a project first", QMessageBox.Ok | QMessageBox.Default, QMessageBox.NoButton
        )
        return

    codes_list = [self.pj[cfg.ETHOGRAM][key][cfg.BEHAVIOR_CODE] for key in self.pj[cfg.ETHOGRAM]]

    self.mapCreatorWindow = BehaviorsMapCreatorWindow(codes_list)
    # behaviors coding map list
    self.mapCreatorWindow.bcm_list = [x["name"].upper() for x in self.pj.get(cfg.BEHAVIORS_CODING_MAP, [])]
    self.mapCreatorWindow.signal_add_to_project.connect(self.behaviors_coding_map_creator_signal_addtoproject)
    self.mapCreatorWindow.move(self.pos())
    self.mapCreatorWindow.resize(cfg.CODING_MAP_RESIZE_W, cfg.CODING_MAP_RESIZE_H)
    self.mapCreatorWindow.show()


if __name__ == "__main__":
    import sys

    app = QApplication(sys.argv)
    window = BehaviorsMapCreatorWindow(["North zone", "East zone", "South zone", "West zone"])
    window.bcm_list = []
    gui_utilities.resize_center(app, window, cfg.CODING_MAP_RESIZE_W, cfg.CODING_MAP_RESIZE_H)
    window.show()
    sys.exit(app.exec())
